# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import dash
from dash import html, dcc, Input, Output, State, ctx
from typing import Optional, Sequence#, Literal

import odatix.components.workspace as workspace
from odatix.gui.icons import icon
from odatix.gui.utils import get_key_from_url
import odatix.gui.ui_components as ui
import odatix.gui.navigation as navigation
import odatix.lib.hard_settings as hard_settings
from odatix.lib.settings import OdatixSettings

page_path = "/jobs_config"

dash.register_page(
    __name__,
    path=page_path,
    title='Odatix - Job Selection',
    name='Jobs',
    order=3,
)

######################################
# UI Components
######################################


######################################
# Callbacks
######################################

@dash.callback(
    Output("job-section", "children"),
    Input("url", "search"),
    State("url", "pathname"),
    State("odatix-settings", "data"),
    prevent_initial_call=True
)
def update_param_domains(
    search, page, odatix_settings
):
    triggered_id = ctx.triggered_id

    if triggered_id == "url":
        if page != page_path:
            return dash.no_update

    arch_path = odatix_settings.get("arch_path", OdatixSettings.DEFAULT_ARCH_PATH)
    architectures = workspace.get_architectures(arch_path)
    job_sections = []
    for arch_name in architectures:     
        domains_configs = {}
        domains = [hard_settings.main_parameter_domain] + workspace.get_param_domains(arch_path, arch_name)
        domain_tiles = []
        for domain in domains:
            if not workspace.check_parameter_domain_use_parameters(arch_path, arch_name, domain):
                continue
            configurations = workspace.get_config_files(arch_path, arch_name, domain)
            if not configurations:
                continue
            configurations = [cfg[:-4] if cfg.endswith('.txt') else cfg for cfg in configurations] # Remove .txt extension
            domains_configs[domain] = configurations
            checklist = dcc.Checklist(
                options=[{"label": cfg, "value": cfg} for cfg in configurations],
                id={"type": "domain-config-checklist", "arch": arch_name, "domain": domain},
                value=configurations,
                style={"width": "max-content", "marginTop": "10px", "marginLeft": "5px", "marginBottom": "10px"},
            )
            domain_tile = html.Div(
                children=[
                    html.Div(
                        children=[
                            html.H3(domain if domain != hard_settings.main_parameter_domain else "Main Parameter Domain", style={"marginBottom": "0px"}),
                            html.Div(
                                children=checklist,
                                style={"overflowX": "scroll", "marginBottom": "-10px"},
                            )
                        ],
                        className="config-domain-content",
                    )
                ],
                className="tile config",
            )
            domain_tiles.append(domain_tile)

        # Default configuration tile
        domain_tiles.append(
             html.Div(
                children=[
                    html.Div(
                        children=[
                            html.H3("Default Configuration", style={"marginBottom": "0px"}),
                            html.Div(
                                children=[
                                    dcc.Checklist(
                                        options=[{"label": f"{arch_name} (default)", "value": arch_name}],
                                        id={"type": "default-config-checklist", "arch": arch_name, "domain": "default"},
                                        value=["default"],
                                        style={"width": "max-content", "marginTop": "10px", "marginLeft": "5px", "marginBottom": "10px"},
                                    ),
                                ],
                                style={"overflowX": "scroll", "marginBottom": "-10px"},
                            )
                        ],
                    )
                ],
                className="tile config",
            )
        )

        all_combinations = [[f"{arch_name}"]] + workspace.generate_config_combinations(domains_configs, arch_name)
        formatted_combinations = [{"label": " + ".join(comb), "value": " + ".join(comb)} for comb in all_combinations]
        # Preview tile
        domain_tiles.append(
            html.Div(
                children=[
                    html.Div(
                        children=[
                            html.H3("Preview", style={"marginBottom": "0px"}),
                            html.Div(
                                children=[
                                    dcc.Checklist(
                                        options=formatted_combinations,
                                        id={"type": "preview-config-checklist", "arch": arch_name},
                                        value=[" + ".join(comb) for comb in all_combinations],
                                        style={"width": "max-content", "marginTop": "10px", "marginLeft": "5px", "marginBottom": "10px"},
                                    )
                                ],
                                style={"overflowX": "scroll", "marginBottom": "-10px"},
                            )
                        ],
                    )
                ],
                className="tile config",
            )
        )
        arch_buttons = html.Div(
            children=[
                ui.icon_button(
                    icon=icon("gear", className="icon"),
                    text="Architecture Settings",
                    color="default",
                    link=f"/arch_editor?arch={arch_name}",
                    multiline=True,
                    width="135px",
                ),
                ui.icon_button(
                    icon=icon("edit", className="icon blue"),
                    text="Edit Configs",
                    tooltip="Open the Configuration Editor for this architecture",
                    tooltip_options="bottom delay",
                    color="default",
                    link=f"/config_editor?arch={arch_name}",
                    multiline=False,
                    width="135px",
                ),
            ],
            className="inline-flex-buttons",
        )            
        job_section = html.Div(
            children=[
                html.Div(
                    children=[
                        ui.title_tile(arch_name, buttons=arch_buttons, id={"type": "arch-title", "arch": arch_name}, switch=False, style={"scale": "1.01"}),
                    ], 
                    id=f"param-domain-title-div-{arch_name}",
                    className="card-matrix config",
                    style={"marginLeft": "-13px"},
                ),
                html.Div(
                    children=domain_tiles,
                    id={"type": "param-domains-container", "arch": arch_name},
                    className="tiles-container config animated-section hide no-margin", 
                    style={"marginBottom": "17px"},
                ),
                dcc.Store(
                    id={"type": "arch-metadata", "arch": arch_name},
                    data={"arch_name": arch_name},
                ),
                dcc.Store(
                    id={"type": "domain-selections", "arch": arch_name},
                    data=domains_configs,
                ),
            ],
            id = {"type": "job-section", "arch": arch_name},
        )
        job_sections.append(job_section)
    return job_sections


@dash.callback(
    Output({"type": "param-domains-container", "arch": dash.ALL}, "className"),                    
    Input({"type": "arch-title", "arch": dash.ALL, "is_switch": True}, "value"),
    State({"type": "arch-metadata", "arch": dash.ALL}, "data"),
)
def toggle_param_domains(selected_archs, arch_metadatas):
    triggered_id = ctx.triggered_id
    if not isinstance(triggered_id, dict):
        return [dash.no_update] * len(selected_archs)
    triggered_arch = triggered_id.get("arch", "")
    updated_classes = []
    for toggled, metadata in zip(selected_archs, arch_metadatas):
        arch_name = metadata.get("arch_name", "")
        if triggered_arch == arch_name:
            if toggled:
                updated_classes.append("tiles-container config animated-section")
            else:
                updated_classes.append("tiles-container config animated-section hide no-margin")
        else:
            updated_classes.append(dash.no_update)
    return updated_classes
    

@dash.callback(
    Output({"type": "domain-selections", "arch": dash.MATCH}, "data"),
    Input({"type": "domain-config-checklist", "arch": dash.MATCH, "domain": dash.ALL}, "value"),
    State({"type": "domain-config-checklist", "arch": dash.MATCH, "domain": dash.ALL}, "id"),
)
def update_domain_selections(selected_per_domain, domain_ids):
    domains_configs = {}
    for values, did in zip(selected_per_domain or [], domain_ids or []):
        domain = did.get("domain")
        if not domain:
            continue
        if values:
            domains_configs[domain] = values
    return domains_configs

@dash.callback(
    Output({"type": "preview-config-checklist", "arch": dash.MATCH}, "value"),
    Input({"type": "domain-config-checklist", "arch": dash.MATCH, "domain": dash.ALL}, "value"),
    State({"type": "domain-config-checklist", "arch": dash.MATCH, "domain": dash.ALL}, "id"),
    State({"type": "preview-config-checklist", "arch": dash.MATCH}, "value"),
    State({"type": "arch-metadata", "arch": dash.MATCH}, "data"),
    State({"type": "domain-selections", "arch": dash.MATCH}, "data"),
)
def sync_preview_values(
    selected_per_domain,
    domain_ids,
    current_preview_values,
    arch_metadata,
    prev_selections
):
    arch_name = arch_metadata.get("arch_name", "")

    # Build the current state of the domains
    current_domains = {}
    for values, did in zip(selected_per_domain or [], domain_ids or []):
        domain = did.get("domain")
        if not domain:
            continue
        if values:
            current_domains[domain] = values

    # Previous state of the domains (can be None on first call)
    prev_domains = prev_selections or {}

    # Find the domain that changed and the added/removed values
    changed_domain = None
    added_values = set()
    removed_values = set()

    # Domains present either before or now
    all_domains = set(prev_domains.keys()) | set(current_domains.keys())
    for domain in all_domains:
        prev_vals = set(prev_domains.get(domain, []))
        curr_vals = set(current_domains.get(domain, []))
        if prev_vals != curr_vals:
            changed_domain = domain
            added_values = curr_vals - prev_vals
            removed_values = prev_vals - curr_vals
            break

    # If no clear change is found, do nothing
    if not changed_domain:
        return current_preview_values or []

    # Start from the current preview value (including manual changes)
    preview_set = set(current_preview_values or [])

    # Helper: generate all complete combinations from current_domains
    all_combos = workspace.generate_config_combinations(current_domains, arch_name)
    all_combo_strings = {" + ".join(c) for c in all_combos}

    # Handle added values in the modified domain
    if added_values:
        for combo in all_combos:
            combo_str = " + ".join(combo)
            # Only handle combos that contain at least one added value
            if any(val in combo_str for val in added_values):
                preview_set.add(combo_str)

    # Handle removed values in the modified domain
    if removed_values:
        to_remove = set()
        for item in preview_set:
            # Do not touch the 'default' item
            if item == arch_name:
                continue
            # If the item is no longer a valid combo, only handle it
            # if it included a removed value from the concerned domain.
            # Simply check if it contains one of the removed values.
            if any(val in item for val in removed_values):
                to_remove.add(item)
        preview_set -= to_remove

    # Keep the "default" item if it was already selected
    if current_preview_values and arch_name in current_preview_values:
        preview_set.add(arch_name)

    # Return a sorted list for display stability
    result = []
    if arch_name in preview_set:
        result.append(arch_name)
        preview_set.remove(arch_name)
    result.extend(sorted(preview_set))

    return result


######################################
# Layout
######################################

layout = html.Div(
    children=[
        dcc.Location(id=f"url"),
        html.Div(id={"page": page_path, "type": "architecture-title-div"}, style={"marginTop": "20px"}),
        html.Div(id="job-section", style={"marginBottom": "10px"}),
    ],
    className="page-content",
    style={
        "display": "flex",  
        "flexDirection": "column",
        "min-height": f"calc(100vh - {navigation.top_bar_height})",
    },
)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import yaml
import re
import csv
import argparse

import odatix.lib.printc as printc
from odatix.lib.utils import read_from_list, create_dir, KeyNotInListError, BadValueInListError
from odatix.lib.get_from_dict import get_from_dict, Key, KeyNotInDictError, BadValueInDictError
import odatix.lib.settings as settings
from odatix.lib.settings import OdatixSettings
from odatix.lib.variables import replace_variables, Variables

current_dir = os.path.dirname(os.path.abspath(__file__))

banned_metrics = []
banned_arch = []

######################################
# Settings
######################################

DEFAULT_FORMAT = "yml"

tool_settings_filename = "tool.yml"

simulations_dir = "simulations"
status_done = "Done: 100%"

script_name = os.path.basename(__file__)


######################################
# Parse Arguments
######################################


def add_arguments(parser):
  """
  Add command-line arguments for configuring the script.

  Args:
      parser (ArgumentParser): Argument parser instance.
  """
  parser.add_argument("-t", "--tool", default="all", help="eda tool in use, or 'all' (default: all)")
  parser.add_argument(
    "-f",
    "--format",
    choices=["csv", "yml", "all"],
    default=DEFAULT_FORMAT,
    help="Output format: csv, yml, or all (default: " + DEFAULT_FORMAT + ")",
  )
  parser.add_argument("-u", "--use_benchmark", action="store_true", help="Use benchmark values in yaml file")
  parser.add_argument("-B", "--benchmark_file", help="Benchmark file")
  parser.add_argument("-w", "--work", help="Work directory")
  parser.add_argument("-r", "--respath", help="Result path")
  parser.add_argument("-m", "--metrics", help="Metrics definition file")
  parser.add_argument(
    "-c",
    "--config",
    default=OdatixSettings.DEFAULT_SETTINGS_FILE,
    help="global settings file for Odatix (default: " + OdatixSettings.DEFAULT_SETTINGS_FILE + ")",
  )


def parse_arguments():
  """
  Parse and return command-line arguments.

  Returns:
      Namespace: Parsed command-line arguments.
  """
  parser = argparse.ArgumentParser(description="Process FPGA or ASIC results")
  add_arguments(parser)
  return parser.parse_args()


######################################
# Parsing functions
######################################


def parse_regex(file, pattern, group_id, error_if_missing=True, error_prefix=""):
  """
  Parse a file with a regex to extract a specific value.

  Args:
      file (str): Path to the file.
      pattern (str): Regular expression to search for.
      group_id (int): Regex group ID to extract.
      error_prefix (str): Prefix for error messages.

  Returns:
      str | None: The matched value, or None if not found or on error.
  """
  if not os.path.isfile(file):
    if error_if_missing:
      printc.error(error_prefix + 'File "' + file + '" does not exist', script_name)
    return None
  with open(file, "r") as f:
    try:
      content = f.read()
      match = re.search(pattern, content)
      if match:
        return match.group(group_id)
    except Exception as e:
      printc.error(
        error_prefix + 'Could not get value from regex "' + pattern + '" in file "' + file + '": ' + str(e), script_name=script_name
      )
      return None

  printc.error(error_prefix + 'No match for regex "' + pattern + '" in file "' + file + '"', script_name=script_name)
  return None


def parse_csv(file, key, error_if_missing=True, error_prefix=""):
  """
  Parse a CSV file to extract a value associated with a specific key.

  Args:
      file (str): Path to the CSV file.
      key (str): Key to search for in the file.
      error_prefix (str): Prefix for error messages.

  Returns:
      str | None: The value corresponding to the key, or None if not found.
  """
  if not os.path.isfile(file):
    if error_if_missing:
      printc.error(error_prefix + 'File "' + file + '" does not exist', script_name)
    return None
  with open(file, mode="r") as csv_file:
    try:
      reader = csv.DictReader(csv_file)
      for row in reader:
        if key in row:
          return row[key]
        else:
          printc.error(error_prefix + 'Could not find key "' + key + '" in csv "' + file + '"', script_name=script_name)
    except csv.Error as e:
      printc.error(error_prefix + 'An error occurred while reading csv file "' + file + '": ' + str(e), script_name=script_name)
      return None

  return None

def parse_yaml(file, key=None, error_if_missing=True, error_prefix=""):
  """
  Parse a YAML file to extract a value associated with a key.

  Args:
      file (str): Path to the YAML file.
      key (str): Key to search for in the YAML data.
      error_prefix (str): Prefix for error messages.

  Returns:
      Any | None: Extracted value, or None if not found or on error.
  """
  if not os.path.isfile(file):
    if error_if_missing:
      printc.error(error_prefix + 'File "' + file + '" does not exist', script_name)
    return None

  with open(file, "r") as yaml_file:
    try:
      data = yaml.safe_load(yaml_file)
      if key:
        value = data.get(key, None)
        if value is None:
          printc.error(error_prefix + 'Could not find key "' + key + '" in yaml "' + file + '"', script_name=script_name)
        return value
      return data
    except yaml.YAMLError as e:
      printc.error(f'{error_prefix}Could not parse yaml file "{file}": {str(e)}')
      return None

  if not os.path.isfile(file):
    printc.error(error_prefix + 'File "' + file + '" does not exist', script_name)
    return None


######################################
# Validate Tool Settings
######################################


def validate_tool_settings(file_path):
  """
  Validate and load tool settings from a YAML file.

  Args:
      file_path (str): Path to the tool settings file.

  Returns:
      dict | None: Loaded settings as a dictionary, or None if invalid.
  """
  if not os.path.isfile(file_path):
    printc.error('Tool settings file "' + os.path.realpath(file_path) + '" does not exist', script_name)
    return None
  with open(file_path, "r") as file:
    try:
      tool_settings = yaml.safe_load(file)
      return tool_settings
    except yaml.YAMLError as e:
      printc.error("Error in tool configuration file: " + str(e), script_name)
      return None


######################################
# Extract Tool Metrics
######################################


def extract_metrics(metrics_data, metrics_file, cur_path, arch, arch_path, use_benchmark, benchmark_file, type="fmax_synthesis"):
  """
  Extract metrics from synthesis results based on tool-specific settings.

  This function processes various types of synthesis metrics, such as 
  `fmax_synthesis` or `custom_freq_synthesis`, by parsing tool output 
  files (e.g., regex, CSV, YAML) and optionally using benchmark values.

  Args:
      metrics_data (dict): The loaded settings for the EDA tool, containing 
                            metric definitions.
      metrics_file (str): Path to the YAML file containing tool settings.
      cur_path (str): Current path to the directory containing synthesis results.
      arch (str): Identifier for the architecture being processed.
      arch_path (str): Full path to the architecture-specific directory.
      use_benchmark (bool): Whether to use benchmark data for extracting metrics.
      benchmark_file (str): Path to the benchmark YAML file.
      type (str): Type of synthesis (e.g., "fmax_synthesis" or "custom_freq_synthesis").
                  Defaults to "fmax_synthesis".

  Returns:
      tuple: 
          - results (dict): A dictionary containing extracted metric values.
            Keys are metric names, and values are the corresponding extracted results.
          - units (dict): A dictionary mapping metric names to their units, if specified.
            If no unit is defined for a metric, it is omitted from this dictionary.

  Raises:
      KeyNotInListError: If a required key is missing in `metrics_data`.
      BadValueInListError: If a value in `metrics_data` is invalid for a metric.
      ValueError: When parsing or formatting a metric value fails.

  Notes:
      - Metrics are extracted from files (regex, CSV, or YAML) defined in the tool 
        settings. Each metric can specify its own type and extraction settings.
      - Metrics marked as "benchmark_only" are only included if `use_benchmark` is True.
      - Global lists `banned_metrics` and `banned_arch` are updated to exclude metrics
        or architectures that encounter errors during extraction.

  Examples:
      For `type="fmax_synthesis"`, metrics defined under "fmax_synthesis_metrics" 
      in the `metrics_data` file are prioritized. Common metrics (defined under 
      "metrics") are always included.
  """
  global banned_metrics
  results = {}
  units = {}
  error_prefix =  arch_path + " => "
  metrics = {}
  
  if type == "fmax_synthesis":
    fmax_metrics = read_from_list("fmax_synthesis_metrics", metrics_data, metrics_file, raise_if_missing=False, print_error=False, script_name=script_name)
    if fmax_metrics != False:
      metrics.update(fmax_metrics)
  elif type == "custom_freq_synthesis":
    range_metrics = read_from_list("custom_freq_synthesis_metrics", metrics_data, metrics_file, raise_if_missing=False, print_error=False, script_name=script_name)
    if range_metrics != False:
      metrics.update(range_metrics)

  common_metrics = read_from_list("metrics", metrics_data, metrics_file, raise_if_missing=False, print_error=False, script_name=script_name)
  if common_metrics != False:
    metrics.update(common_metrics)

  for metric, content in metrics.items():
    if metric in banned_metrics:
      continue

    try:
      type = read_from_list("type", content, metrics_file, parent=metric, script_name=script_name)
      settings = read_from_list("settings", content, metrics_file, parent=metric, script_name=script_name)
    except (KeyNotInListError, BadValueInListError):
      banned_metrics.append(metric)
      continue

    benchmark_only = read_from_list("benchmark_only", content, metrics_file, parent=metric, raise_if_missing=False, type=bool, print_error=False, script_name=script_name)
    if benchmark_only and not use_benchmark:
      banned_metrics.append(metric)
      continue

    error_if_missing, _ = get_from_dict("error_if_missing", content, metrics_file, parent=metric, default_value=True, type=bool, silent=True, script_name=script_name)

    if type == "regex":
      try:
        file = read_from_list("file", settings, metrics_file, parent=metric + "[settings]", script_name=script_name)
        pattern = read_from_list("pattern", settings, metrics_file, parent=metric + "[settings]", script_name=script_name)
        group_id = read_from_list("group_id", settings, metrics_file, parent=metric + "[settings]", type=int, script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        banned_metrics.append(metric)
        continue
      value = parse_regex(os.path.join(cur_path, file), pattern, group_id, error_if_missing, error_prefix)
    elif type == "csv":
      try:
        file = read_from_list( "file", settings, metrics_file, parent=metric + "[settings]", script_name=script_name)
        key = read_from_list("key", settings, metrics_file, parent=metric + "[settings]", script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        banned_metrics.append(metric)
        continue
      value = parse_csv(os.path.join(cur_path, file), key, error_if_missing, error_prefix)
    elif type == "yaml":
      try:
        file = read_from_list("file", settings, metrics_file, parent=metric + "[settings]", script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        banned_metrics.append(metric)
        continue
      key, _ = get_from_dict("key", settings, metrics_file, parent=metric + "[settings]", silent=True, default_value=None, script_name=script_name)
      value = parse_yaml(os.path.join(cur_path, file), key, error_if_missing, error_prefix)
    elif type == "benchmark":
      if not use_benchmark:
        banned_metrics.append(metric)
        continue
      if arch in banned_arch:
        continue
      try:
        key = read_from_list("key", settings, metrics_file, parent=metric + "[settings]", script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        banned_metrics.append(metric)
        continue
      key = arch + "[" + key + "]"
      value = parse_yaml(benchmark_file, key, error_if_missing, error_prefix)
      if value is None:
        banned_arch.append(arch)
    elif type == "operation":
      try:
        op = read_from_list("op", settings, metrics_file, parent=metric + "[settings]", script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        banned_metrics.append(metric)
        continue
      value = calculate_operation(op, results, error_if_missing, error_prefix)
    else:
      printc.error(
        'Unsupported metric type "' + type + '" specified for metric "' + metric + '" in "' + metrics_file + '"',
        script_name=script_name,
      )
      banned_metrics.append(metric)
      continue

    # Apply formatting if specified
    if value is not None and "format" in content:
      try:
        value = convert_to_numeric(content["format"] % float(value))
      except ValueError:
        pass  # printc.warning(f"Failed to format value {value} for metric {metric}", script_name)

    # Append unit if specified
    if value is not None:
      results[metric] = value
      if "unit" in content:
        units[metric] = content["unit"]
    else:
      results[metric] = None

  return results, units


######################################
# Misc functions
######################################


def corrupted_directory(directory):
  """
  Log a warning indicating that a synthesis directory is corrupted or incomplete.

  Args:
      directory (str): Path to the directory that is flagged as corrupted.
  """
  printc.warning(
    directory + " => Synthesis has not finished or directory has been corrupted", script_name
  )


def convert_to_numeric(data):
  """
  Convert a string representation of a number to a numeric type (int or float).

  Args:
      data (str): The string to convert.

  Returns:
      int | float | str: The numeric representation of the input if conversion is successful;
      otherwise, the original string.
  """
  try:
    if "." in data:
      return float(data)
    return int(data)
  except ValueError:
    return data


def calculate_operation(op_str, results, error_if_missing=True, error_prefix=""):
  """
  Evaluate a mathematical operation on the extracted results.

  Args:
      op_str (str): The mathematical operation as a string (e.g., "a + b / c").
      results (dict): Dictionary of metric values available for the operation.
      error_prefix (str): Prefix for error messages to provide context.

  Returns:
      float | int | None: The result of the evaluated operation, or None if an error occurs.
  """
  try:
    local_vars = {k: v for k, v in results.items() if v is not None}
    return eval(op_str, {}, local_vars)
  except (NameError, SyntaxError, TypeError, ZeroDivisionError) as e:
    if error_if_missing:
      printc.error(error_prefix + 'Failed to evaluate operation "' + op_str + '": ' + str(e) , script_name)
    return None


######################################
# Export Results
######################################

def process_configuration(input, target, architecture, configuration, frequency, type, result_key, units, data, metrics_data, metrics_file, use_benchmark, benchmark_file):
  """
  Process the configuration for a specific architecture and extract relevant metrics.

  This function validates the synthesis status for a given configuration and 
  architecture, and extracts metrics if the synthesis has completed successfully.

  Args:
      input (str): Base directory for the synthesis results.
      target (str): Target device or platform being synthesized.
      architecture (str): Name of the architecture being processed.
      configuration (str): Configuration of the architecture (e.g., specific design options).
      frequency (str | None): Frequency variant for custom frequency synthesis; None for fmax synthesis.
      type (str): Type of results to process (e.g. "fmax_synthesis" or "custom_freq_synthesis").
      units (dict): Dictionary to store metric units for the synthesis results.
      data (dict): Dictionary to store extracted metrics for the current configuration.
      metrics_data (dict): Settings specific to the EDA tool being used.
      metrics_file (str): Path to the YAML file containing tool settings.
      use_benchmark (bool): Whether to use benchmark values for metric extraction.
      benchmark_file (str): Path to the benchmark YAML file.

  Returns:
      tuple:
          - data (dict): Updated dictionary containing extracted metrics.
          - metrics (dict): Extracted metrics for the current configuration.

  Notes:
      - This function ensures the synthesis status is checked before attempting 
        to extract metrics.
      - For `custom_freq_synthesis`, metrics are organized per frequency.
      - Updates the global `units` dictionary with units for the extracted metrics.

  Raises:
      None: Errors are logged using `printc.error`, and the function returns None if an error occurs.
  """
  if type == "custom_freq_synthesis":
    arch = architecture + "[" + configuration + "] @ " + frequency
    arch_path = os.path.join(target, architecture, configuration, frequency)
    status_filename = "synth_status.log"
  else:
    arch = architecture + "[" + configuration + "]"
    arch_path = os.path.join(target, architecture, configuration)
    status_filename = "status.log"

  cur_path = os.path.join(input, arch_path)
  status_log = os.path.join(cur_path, "log", status_filename)

  metrics = {}
  cur_units = {}

  # Check if synthesis completed
  if not os.path.isfile(status_log):
    corrupted_directory(arch_path)
    return None, None

  with open(status_log, "r") as f:
    if status_done not in f.read():
      corrupted_directory(arch_path)
      return None, None

  # Get values
  if type == "custom_freq_synthesis":
    metrics[result_key], cur_units[result_key] = extract_metrics(metrics_data, metrics_file, cur_path, arch, arch_path, use_benchmark, benchmark_file, type)
    data[result_key][target][architecture][configuration][frequency] = metrics[result_key]
  else:
    metrics[result_key], cur_units[result_key] = extract_metrics(metrics_data, metrics_file, cur_path, arch, arch_path, use_benchmark, benchmark_file, type)
    data[result_key][target][architecture][configuration] = metrics[result_key]
  # print(f"metrics[result_key] = {metrics[result_key]}")

  # Update units
  units.update(cur_units[result_key])
  return data, metrics


def export_results(input, output, tools, format, use_benchmark, benchmark_file, result_types, custom_metrics_file=None):
  """
  Export synthesis results for multiple tools, configurations and architectures
  to a specified format.

  This function iterates over synthesis results from specified tools, of every
  architecture configurations in the input directory, extracts metrics, and writes 
  the results to YAML files.

  Args:
      input (str): Base directory containing synthesis results.
      output (str): Directory where the exported results will be saved.
      tools (list | str): List of EDA tools to process or "All" to process all available tools.
      format (str): Output format for the results ("csv", "yml", or "all").
      use_benchmark (bool): Whether to include benchmark data in the metrics.
      benchmark_file (str): Path to the benchmark YAML file.
      result_types (dict): Dictionary specifying the yaml key and path for each result type to process (e.g., `fmax_synthesis` and `custom_freq_synthesis`) 
      custom_metrics_file (str): Path to a metrics definition YAML file.

  Returns:
      None: Results are saved to files; errors are logged.

  Notes:
      - For each tool, synthesis types (`fmax_synthesis` and `custom_freq_synthesis`) 
        are processed separately.
      - Creates a structured YAML output containing extracted metrics and units.
      - Skips tools or configurations with missing or invalid data.
      - Outputs results to separate YAML files for each tool.

  Example:
      If `tools = ["tool1", "tool2"]`, the function generates:
      - `results_tool1.yml` in the specified output directory.
      - `results_tool2.yml` in the specified output directory.

  Raises:
      SystemExit: If no valid synthesis results are found for the specified tools.
  """
  input_path = input

  # Check result_types is valid
  for result_type in result_types:
    if not "path" in result_types[result_type] or not "key" in result_types[result_type]:
      printc.error("Invalid result_types formatting: " + str(result_types), script_name)
      printc.note("Example of valid synthesis type formatting: ", script_name)
      printc.cyan('"custom_freq_synthesis": {')
      printc.cyan('  "key": "custom_freq_synthesis",')
      printc.cyan('  "path": your/custom/freq/synthesis/path,')
      printc.cyan('},')
      printc.cyan('"fmax_synthesis": {')
      printc.cyan('  "key": "fmax_synthesis",')
      printc.cyan('  "path": your/fmax/synthesis/path,')
      printc.cyan('},: ')
      return

  # Get tool list
  if not isinstance(tools, list):
    if not isinstance(tools, str) or tools != "all":
      printc.error("Invalid value for 'tools': " + tools, script_name)
      printc.note("'tools' should be ether a list or 'all'" + tools, script_name)
    else:
      tools = []
      for result_type in result_types:
        work_path = result_types[result_type]["path"]
        type_dir = os.path.join(input_path, work_path)
        if os.path.isdir(type_dir):
          tools += [item for item in os.listdir(type_dir) if os.path.isdir(os.path.join(input, result_type, item))]
      tools = list(set(tools))

  for tool in tools:
    data = {}
    units = {}
    cur_units = {}
    metrics = {}

    # Get tool setting file
    tool_settings_file = os.path.join(OdatixSettings.odatix_eda_tools_path, tool, tool_settings_filename)
    tool_settings = validate_tool_settings(tool_settings_file)
    if tool_settings is None:
      if len(tools) == 1:
        sys.exit(-1)
      else:
        continue

    # Get the default_metrics_file from the tool settings file
    if custom_metrics_file is None:
      metrics_file, defined = get_from_dict("default_metrics_file", tool_settings, tool_settings_file, behavior=Key.MANTADORY, script_name=script_name)
      if not defined:
        continue
    else:
      metrics_file = custom_metrics_file

    # Define user accessible variables
    variables = Variables(
      odatix_path=OdatixSettings.odatix_path,
      odatix_eda_tools_path=OdatixSettings.odatix_eda_tools_path,
    )

    # Replace variables in command
    metrics_file = replace_variables(metrics_file, variables)

    if not os.path.isfile(metrics_file):
      printc.error('Metrics definition file "' + os.path.realpath(metrics_file) + '" does not exist', script_name)
      continue
    with open(metrics_file, "r") as file:
      try:
        metrics_data = yaml.safe_load(file)
      except yaml.YAMLError as e:
        printc.error("Error in metrics definition file: " + str(e), script_name)
        continue

    for result_type in result_types:
      result_key = result_types[result_type]["key"]
      work_path = result_types[result_type]["path"]
      data[result_key] = {}
      printc.cyan("Export " + tool + " " + result_key + " results", script_name)

      input = os.path.join(input_path, work_path, tool)

      try:
        dirs = sorted(next(os.walk(input))[1])
      except StopIteration:
        continue

      for target in dirs:
        data[result_key][target] = {}
        for architecture in sorted(next(os.walk(os.path.join(input, target)))[1]):
          data[result_key][target][architecture] = {}
          for configuration in sorted(next(os.walk(os.path.join(input, target, architecture)))[1]):
            if result_type == "custom_freq_synthesis":
              data[result_key][target][architecture][configuration] = {}
              for frequency in sorted(next(os.walk(os.path.join(input, target, architecture, configuration)))[1]):
                process_configuration(input, target, architecture, configuration, frequency, result_type, result_key, units, data, metrics_data, metrics_file, use_benchmark, benchmark_file)
              if data == None:
                continue
            else:
              process_configuration(input, target, architecture, configuration, None, result_type, result_key, units, data, metrics_data, metrics_file, use_benchmark, benchmark_file)
            
            if data == None:
              continue

    # Export to the desired format
    os.makedirs(output, exist_ok=True)
    output_file = os.path.join(output, "results_" + tool + ".yml")
    try:
      with open(output_file, "w") as file:
        yaml.dump(
          {"units": units, "fmax_synthesis": data["fmax_synthesis"], "custom_freq_synthesis": data["custom_freq_synthesis"]}, file, default_style=None, default_flow_style=False, sort_keys=False
        )
        printc.say('Results written to "' + output_file + '"', script_name=script_name)
        printc.note("Run 'odatix-explorer' to explore the results", script_name=script_name)
    except Exception as e:
      printc.error('Could not write "' + output_file + '"', script_name=script_name)
      printc.cyan("error details: ", script_name=script_name, end="")
      print(str(e))


######################################
# Main
######################################


def main(args, settings=None):
  # Get settings
  if settings is None:
    settings = OdatixSettings(args.config)
    if not settings.valid and (args.work is None or args.respath is None):
      printc.error("Could not load settings from file \"" + args.config + "\" and -w and/or -r options are not used", script_name=script_name)
      sys.exit(-1)

  if args.use_benchmark is not None:
    use_benchmark = args.use_benchmark
  else:
    if settings.valid:
      use_benchmark = settings.use_benchmark
    else:
      use_benchmark = False
      benchmark_file = None

  if args.benchmark_file is not None:
    benchmark_file = args.benchmark_file
  else:
    if settings.valid:
      benchmark_file = settings.benchmark_file
    else:
      use_benchmark = False
      benchmark_file = None

  if args.work is not None:
    input = args.work
  else:
    input = settings.work_path

  if not os.path.isdir(input):
    printc.error('Could not find work directory "' + input + '"', script_name=script_name)
    printc.note("Run fmax synthesis using the 'odatix fmax' command before exporting the results", script_name=script_name)
    printc.note("Or run custom frequency synthesis using the 'odatix freq' command before exporting the results", script_name=script_name)
    sys.exit(-1)

  if args.respath is not None:
    output = args.respath
  else:
    output = settings.result_path

  if settings.valid:
    result_types = settings.result_types
  else:
    result_types =  {
      "custom_freq_synthesis": {
        "key": "custom_freq_synthesis",
        "path": OdatixSettings.DEFAULT_CUSTOM_FREQ_SYNTHESIS_WORK_PATH
      },
      "fmax_synthesis": {
        "key": "fmax_synthesis",
        "path": OdatixSettings.DEFAULT_FMAX_SYNTHESIS_WORK_PATH
      },
    }

  if args.tool == "all":
    tools = args.tool
  else:
    tools = [args.tool]

  if args.metrics is not None:
    metrics_file = args.metrics
  else:
    metrics_file = None

  export_results(
    input=input,
    output=output,
    tools=tools,
    format=args.format,
    use_benchmark=use_benchmark,
    benchmark_file=benchmark_file,
    result_types=result_types,
    custom_metrics_file=metrics_file,
  )


if __name__ == "__main__":
  args = parse_arguments()
  main(args)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import dash
from dash import html, dcc, Input, Output, State, ctx
from typing import Optional#, Literal

import odatix.gui.navigation as navigation
import odatix.gui.ui_components as ui
from odatix.gui.css_helper import Style
from odatix.gui.utils import get_key_from_url
from odatix.gui.icons import icon
from odatix.lib.settings import OdatixSettings 
import odatix.lib.hard_settings as hard_settings
from odatix.lib.config_generator import ConfigGenerator
import odatix.components.workspace as workspace

verbose = False

page_path = "/config_generator"

dash.register_page(
    __name__,
    path=page_path,
    title='Odatix - Configuration Generator',
    name='Configuration Generator',
    order=5,
)

######################################
# Conversion Functions
######################################

def get_gen_settings(
    name, template,
    titles, types, base_vals, from_vals, to_vals, from_2_pow_vals, to_2_pow_vals, from_type_vals, to_type_vals, step_vals, op_vals, list_vals, source_vals, sources_vals, format_vals
):
    """
    Get the configuration generation settings from the UI inputs.
    Args:
        name (str): The name of the configuration.
        template (str): The template for the configuration.
        titles (list): List of variable names.
        types (list): List of variable types.
        base_vals (list): List of base values for "multiples" type.
        from_vals (list): List of from values for "range" and "multiples" types.
        to_vals (list): List of to values for "range" and "multiples" types.
        from_2_pow_vals (list): List of from 2^ values for "power_of_two" type.
        to_2_pow_vals (list): List of to 2^ values for "power_of_two" type.
        from_type_vals (list): List of from values for "conversion" type.
        to_type_vals (list): List of to values for "conversion" type.
        step_vals (list): List of step values for "range" type.
        op_vals (list): List of operation strings for "function" type.
        list_vals (list): List of comma-separated values for "list" type.
        source_vals (str): Dource variable name for "conversion" type.
        sources_vals (list): List of comma-separated source variable names for set operations.
        format_vals (list): List of format strings.
    Returns:
        dict: Configuration generation settings.
    """
    variables = {}
    for idx, title in enumerate(titles):
        type = types[idx]
        settings = {}
        format = format_vals[idx] if format_vals[idx] else None
        if type == "range":
            settings["from"] = int(from_vals[idx]) if from_vals[idx] else 0
            settings["to"] = int(to_vals[idx]) if to_vals[idx] else 0
            settings["step"] = int(step_vals[idx]) if step_vals[idx] else 1
        elif type == "power_of_two":
            settings["from_2^"] = int(from_2_pow_vals[idx]) if from_2_pow_vals[idx] else 0
            settings["to_2^"] = int(to_2_pow_vals[idx]) if to_2_pow_vals[idx] else 0
        elif type == "list":
            settings["list"] = [x.strip() for x in list_vals[idx].split(",") if x.strip()] if list_vals[idx] else []
        elif type == "multiples":
            settings["base"] = int(base_vals[idx]) if base_vals[idx] else 1
            settings["from"] = int(from_vals[idx]) if from_vals[idx] else 0
            settings["to"] = int(to_vals[idx]) if to_vals[idx] else 0
        elif type == "function":
            settings["op"] = op_vals[idx] if op_vals[idx] else ""
        elif type == "conversion":
            settings["from"] = from_type_vals[idx] if from_type_vals[idx] else 0
            settings["to"] = to_type_vals[idx] if to_type_vals[idx] else 0
            settings["source"] = source_vals[idx] if source_vals[idx] else ""
        elif type == "format":
            settings["source"] = source_vals[idx] if source_vals[idx] else ""
        elif type in {"union", "disjunctive_union", "intersection", "difference"}:
            settings["sources"] = [x.strip() for x in sources_vals[idx].split(",") if x.strip()] if sources_vals[idx] else []
        variable = workspace.create_config_gen_variable_dict(name=title, type=type, settings=settings, format=format)
        variables.update(variable)
    gen_settings = workspace.create_config_gen_dict(name=name, template=template, variables=variables)
    return gen_settings

######################################
# UI Components
######################################

def config_parameters_form(settings):
    defval = lambda k, v=None: settings.get(k, v)
    return html.Div(
        children=[
            html.H3(f"Configuration Generation Settings"),
            html.Div([
                html.Label("Configuration name"),
                ui.tooltip_icon("The name template for the configurations to be generated. The extension '.txt' is not required. Use variable names enclosed in curly braces, prefixed by a dollar sign (e.g., ${variable_name}) to indicate where variable values should be inserted."),
                dcc.Input(
                    id="generator-name",
                    value=defval("name", ""),
                    type="text",
                    style={"width": "100%", "fontFamily": "monospace", "fontWeight": "500"}
                ),
            ], style={"marginBottom": "12px"}),
            html.Div([
                html.Label("Content template"),
                ui.tooltip_icon("The content template for the configurations to be generated. Use variable names enclosed in curly braces, prefixed by a dollar sign (e.g., ${variable_name}) to indicate where variable values should be inserted."),
                dcc.Textarea(
                    id="generator-template",
                    value=defval("template", ""),
                    className="auto-resize-textarea",
                    style={"width": "100%", "resize": "none", "fontFamily": "monospace", "fontWeight": "500"},
                ),
            ], style={"marginBottom": "12px"}),
            dcc.Store(id="generator-initial-settings", data=settings),
            dcc.Store(id="generator-saved-settings", data=None),
        ],
        id="config-parameters",
        className="tile config",
    )

def variable_field(
    var: str,
    # type: Literal['text', 'number', 'password', 'email', 'range', 'search', 'tel', 'url', 'hidden'] = "text",
    type = "text",
    name: str = "",
    label: Optional[str] = None,
    options: Optional[list] = None,
    value: str = "",
    placeholder: str = "",
    default_style: dict = Style.hidden
):
    if label is None:
        label = name
    return html.Div(
        children=[
            html.Div(
                children=[
                    html.Label(label, style={"fontWeight": "bold", "fontSize": "1em"}),
                    dcc.Input(
                        value=value,
                        type=type,
                        placeholder=placeholder,
                        id={"type": f"variable-field-{name}", "name": var},
                        className="value-input",
                        style={
                            "width": "calc(100% - 20px)",
                            "marginLeft": "5px",
                            "marginRight": "5px",
                            "marginBottom": "5px",
                            "fontSize": "0.9em",
                            "height": "10px",
                            "zIndex": "900",
                        },
                    ) if options is None else dcc.Dropdown(
                        id={"type": f"variable-field-{name}", "name": var},
                        options=options,
                        value=value,
                        clearable=False,
                        style={
                            "fontSize": "0.95em",
                            "zIndex": "900",
                        },
                    ),
                ], 
                style={"marginTop": "5px"}
            ),
        ],
        id={"type": f"variable-field-{name}-div", "name": var},
        style=default_style
    )

def variable_card(
        name, var_layout="normal",
        type_value="list",
        base_value="", from_value="", to_value="",
        from_2_pow_value="", to_2_pow_value="", step_value="1",
        from_type_value="dec", to_type_value="hex",
        op_value="", list_value="", source_value="", sources_value="", format_value="",
    ):
    save_class =  "color-button disabled"
    return html.Div([
        html.Div([
            dcc.Input(
                value=name,
                type="text",
                id={"type": "variable-title", "name": name},
                className="title-input",
                style={
                    "width": "calc(100% - 20px)",
                    "marginLeft": "5px",
                    "marginRight": "5px",
                    "marginTop": "-5px",
                    "marginBottom": "2px",
                    "fontWeight": "bold",
                    "fontSize": "1.1em",
                    "height": "10px",
                    "textAlign": "center",
                },
            ),
            dcc.Dropdown(
                id={"type": "variable-type", "name": name},
                options=[
                    {"label": "Boolean", "value": "bool"},
                    {"label": "List", "value": "list"},
                    {"label": "Range", "value": "range"},
                    {"label": "Power of 2", "value": "power_of_two"},
                    {"label": "Multiples", "value": "multiples"},
                    {"label": "Function", "value": "function"},
                    {"label": "Conversion", "value": "conversion"},
                    {"label": "Format", "value": "format"},
                    {"label": "Union", "value": "union"},
                    {"label": "Disjunctive Union", "value": "disjunctive_union"},
                    {"label": "Intersection", "value": "intersection"},
                    {"label": "Difference", "value": "difference"},
                ],
                value=type_value,
                clearable=False,
            ),
            html.Div(
                children=[ 
                    variable_field(var=name, name="from", label="From", type="number", value=from_value),
                    variable_field(var=name, name="to", label="To", type="number", value=to_value),
                    variable_field(var=name, name="from_2_pow", label="From 2^", type="number", value=from_2_pow_value),
                    variable_field(var=name, name="to_2_pow", label="To 2^", type="number", value=to_2_pow_value),
                    variable_field(var=name, name="from_type", label="From type", type="text", options=[{"label": "Bin", "value": "bin"}, {"label": "Dec", "value": "dec"}, {"label": "Hex", "value": "hex"}], value=from_type_value),
                    variable_field(var=name, name="to_type", label="To type", type="text", options=[{"label": "Bin", "value": "bin"}, {"label": "Dec", "value": "dec"}, {"label": "Hex", "value": "hex"}], value=to_type_value),
                    variable_field(var=name, name="base", label="Base", type="number", value=base_value),
                    variable_field(var=name, name="step", label="Step", type="number", value=step_value),
                    variable_field(var=name, name="op", label="Op", type="text", value=op_value),
                    variable_field(var=name, name="list", label="List", type="text", placeholder="Comma-separated values", default_style=Style.visible, value=list_value),
                    variable_field(var=name, name="source", label="Source", type="text", value=source_value),
                    variable_field(var=name, name="sources", label="Sources", type="text", placeholder="Comma-separated values", value=sources_value),
                    html.Div(
                        children=[
                            variable_field(var=name, name="format", label="Format", type="text", value=format_value),
                        ],
                        id={"type": "more-variable-field-div", "name": name},
                        className="expandable-area",
                        style=Style.hidden
                    )
                ],
                id="variable-fields-container",
            ),
        ]),
        html.Div([
            html.Div([
                ui.icon_button(
                    icon=icon("more", className="icon normal rotate", id={"type": "more-fields-icon", "name": name}),
                    color="default",
                    id={"type": "more-fields", "name": name},
                    tooltip="Show/Hide extra fields",
                    tooltip_options="bottom small",
                )
            ], id={"type": "more-fields-div", "name": name}, style={"display": "flex", "alignItems": "center"}),
            html.Div([
                ui.duplicate_button(id={"type": "duplicate-var", "name": name}),
                ui.delete_button(id={"type": "delete-var", "name": name}),
            ], style={"display": "flex", "flexDirection": "hotizontal", "alignItems": "center"}),
        ], style={
            "marginTop": "8px",
            "display": "flex",
            "flexDirection": "row",
            "width": "100%",
            "justifyContent": "space-between",
        }),
        dcc.Store(id={"type": "variable-metadata", "name": name}, data={"name": name, "type": type_value, "base_value": base_value, "from_value": from_value, "to_value": to_value, "from_2_pow_value": from_2_pow_value, "to_2_pow_value": to_2_pow_value, "from_type_value": from_type_value, "to_type_value": to_type_value, "step_value": step_value, "op_value": op_value, "list_value": list_value, "source_value": source_value, "sources_value": sources_value, "format_value": format_value}),
    ], 
    className="card configs", 
    id={"type": "variable-card", "name": name},
    style={
        "padding": "10px", 
        "margin": "5px", 
        "display": "inline-block", 
        "verticalAlign": "top"
    })

def add_card(text: str = "Add new variable"):
    return html.Div(
        html.Div(
            html.Div(
                children=[
                    html.Div(text, style={"fontWeight": "bold", "fontSize": "1.2em", "paddingTop": "20px"}),
                    html.Div(
                        "+",
                        style={
                            "fontSize": "2.5em",
                            "lineHeight": "80px",
                            "height": "80px",
                        }
                    ),
                ], 
                style={"display": "flex", "flexDirection": "column", "alignItems": "center", "justifyContent": "center", "height": "100%"}
            ),
            id="new-variable",
            n_clicks=0,
            style={"textDecoration": "none", "height": "100%"},
        ),
        className=f"card configs add hover",
        id="add-config-card",
        style={
            "padding": "10px",
            "margin": "5px",
            "display": "inline-block",
            "verticalAlign": "top",
            "boxSizing": "border-box"
        },
    )

def config_card(filename, content, config_layout="normal", ellipsis=False):
    display_name = filename[:-4] if filename.endswith(".txt") else filename
    return html.Div([
        html.Div([
            dcc.Input(
                value=f"{display_name}",
                type="text",
                id={"type": "config-title", "filename": filename},
                className="title-input preview",
                readOnly=True,
                style={
                    "width": "calc(100% - 20px)",
                    "marginLeft": "5px",
                    "marginRight": "5px",
                    "fontWeight": "bold",
                    "fontSize": "1.1em",
                    "height": "10px",
                    "marginTop": "-5px",
                    "marginBottom": "2px",
                    "textAlign": "center",
                },
            )
        ]) if not ellipsis else html.Div(style={"height": "24px"}),
        dcc.Textarea(
            id={"type": "config-content", "filename": filename},
            value=content,
            className="auto-resize-textarea" if config_layout != "compact" else "",
            readOnly=True,
            style={
                "width": "calc(100% - 20px)",
                "marginLeft": "5px",
                "marginRight": "5px",
                "resize": "none" if config_layout != "compact" else "vertical",
                "minHeight": "none" if config_layout != "compact" else "45px",
                "height": "none" if config_layout != "compact" else "45px",
                "fieldSizing": "border-box",
                "fontFamily": "monospace",
                "fontSize": "0.9em",
                "fontWeight": "normal",
            },
        ) if not ellipsis else html.Div(
            children=[
                html.Span('[...]', style={"margin": "0", "padding": "0"}),
                html.H4(content, style={"margin": "0", "padding": "0"}), 
            ],
            style={
                "display": "flex",
                "flexDirection": "column",
                "alignItems": "center",
                "justifyContent": "center",
                "height": "45px",
                "gap": "0px",
            },        
        ), 
    ],
    className="card configs", 
    id={"type": "config-card", "filename": filename},
    style={
        "padding": "10px", 
        "margin": "5px", 
        "display": "inline-block", 
        "border": "1px dashed #bbb" if ellipsis else "",
        "verticalAlign": "top"
    })


######################################
# Callbacks
######################################

@dash.callback(
    [
        Output({"type": "variable-field-from-div", "name": dash.ALL}, "style"),
        Output({"type": "variable-field-to-div", "name": dash.ALL}, "style"),
        Output({"type": "variable-field-from_2_pow-div", "name": dash.ALL}, "style"),
        Output({"type": "variable-field-to_2_pow-div", "name": dash.ALL}, "style"),
        Output({"type": "variable-field-from_type-div", "name": dash.ALL}, "style"),
        Output({"type": "variable-field-to_type-div", "name": dash.ALL}, "style"),
        Output({"type": "variable-field-base-div", "name": dash.ALL}, "style"),
        Output({"type": "variable-field-step-div", "name": dash.ALL}, "style"),
        Output({"type": "variable-field-op-div", "name": dash.ALL}, "style"),
        Output({"type": "variable-field-list-div", "name": dash.ALL}, "style"),
        Output({"type": "variable-field-source-div", "name": dash.ALL}, "style"),
        Output({"type": "variable-field-sources-div", "name": dash.ALL}, "style"),
        Output({"type": "variable-field-format-div", "name": dash.ALL}, "style"),
    ],
    Input({"type": "variable-type", "name": dash.ALL}, "value"),
)
def update_variable_fields_visibility(types):
    # Required fields for each type
    mapping = {
        "bool":              {"format"},
        "list":              {"list", "format"},
        "range":             {"from", "to", "step", "format"},
        "power_of_two":      {"from_2_pow", "to_2_pow", "format"},
        "multiples":         {"base", "from", "to", "format"},
        "function":          {"op", "format"},
        "conversion":        {"from_type", "to_type", "source", "format"},
        "format":            {"source", "format"},
        "union":             {"sources", "format"},
        "disjunctive_union": {"sources", "format"},
        "intersection":      {"sources", "format"},
        "difference":        {"sources", "format"},
    }
    all_fields = ["from", "to", "from_2_pow", "to_2_pow", "from_type", "to_type", "base", "step", "op", "list", "source", "sources", "format"]

    styles_by_field = {field: [] for field in all_fields}
    for t in types:
        visible = mapping.get(t, set())
        for field in all_fields:
            styles_by_field[field].append(Style.visible if field in visible else Style.hidden)

    return (
        styles_by_field["from"],
        styles_by_field["to"],
        styles_by_field["from_2_pow"],
        styles_by_field["to_2_pow"],
        styles_by_field["from_type"],
        styles_by_field["to_type"],
        styles_by_field["base"],
        styles_by_field["step"],
        styles_by_field["op"],
        styles_by_field["list"],
        styles_by_field["source"],
        styles_by_field["sources"],
        styles_by_field["format"],
    )

@dash.callback(
    Output({"type": "variable-cards-row"}, "children"),
    Output("generator-name", "value"),
    Output("generator-template", "value"),
    Output("generator-initial-settings", "data"),
    Input("url", "search"),
    Input("url", "pathname"),
    Input("new-variable", "n_clicks"),
    Input({"type": "duplicate-var", "name": dash.ALL}, "n_clicks"),
    Input({"type": "delete-var", "name": dash.ALL}, "n_clicks"),
    State({"type": "variable-cards-row"}, "children"),
    State({"type": "variable-type", "name": dash.ALL}, "value"),
    State({"type": "variable-field-base", "name": dash.ALL}, "value"),
    State({"type": "variable-field-from", "name": dash.ALL}, "value"),
    State({"type": "variable-field-to", "name": dash.ALL}, "value"),
    State({"type": "variable-field-from_2_pow", "name": dash.ALL}, "value"),
    State({"type": "variable-field-to_2_pow", "name": dash.ALL}, "value"),
    State({"type": "variable-field-from_type", "name": dash.ALL}, "value"),
    State({"type": "variable-field-to_type", "name": dash.ALL}, "value"),
    State({"type": "variable-field-step", "name": dash.ALL}, "value"),
    State({"type": "variable-field-op", "name": dash.ALL}, "value"),
    State({"type": "variable-field-list", "name": dash.ALL}, "value"),
    State({"type": "variable-field-source", "name": dash.ALL}, "value"),
    State({"type": "variable-field-sources", "name": dash.ALL}, "value"),
    State({"type": "variable-field-format", "name": dash.ALL}, "value"),
    State("odatix-settings", "data"),
    prevent_initial_call=True
)
def update_form_and_variable_cards(
    search, page, new_click, duplicate_clicks, delete_clicks, cards,
    types, base_vals, from_vals, to_vals, from_2_pow_vals, to_2_pow_vals, from_type_vals, to_type_vals, step_vals, op_vals, list_vals, source_vals, sources_vals, format_vals,
    odatix_settings
):
    trigger_id = ctx.triggered_id

    if trigger_id == "url":
        if page != page_path:
            return dash.no_update, dash.no_update, dash.no_update, dash.no_update

        arch_path = odatix_settings.get("arch_path", OdatixSettings.DEFAULT_ARCH_PATH)
        arch_name = get_key_from_url(search, "arch")
        domain = get_key_from_url(search, "domain")
        if not domain: 
            domain = hard_settings.main_parameter_domain
        if not arch_name:
            return [], dash.no_update, dash.no_update, dash.no_update

        settings = workspace.load_architecture_settings(arch_path, arch_name, domain)
        variables = {}

        generator_name = ""
        generator_template = ""
        gen_settings = {}
        if "generate_configurations_settings" in settings:
            gen_settings = settings["generate_configurations_settings"]
            variables = gen_settings.get("variables", {})
            generator_name = gen_settings.get("name", "")
            generator_template = gen_settings.get("template", "")
            if isinstance(generator_template, list):
                generator_template = "\n".join(generator_template)
        else:
            # Default template
            generator_name = "config_${var1}${var2}"
            generator_template = "parameter VALUE1 = ${var1};\nparameter VALUE2 = ${var2};"
            variables = {
                "var1": {"type": "range", "settings": {"from": 1, "to": 3, "step": 1}},
                "var2": {"type": "list", "settings": {"list": ["A", "B", "C", "D"]}},
            }
        
        cards = []
        # Create cards from existing variables
        for var_name, var_keys in variables.items():
            type_value = var_keys.get("type", "list")
            var_settings = var_keys.get("settings", {})
            card_kwargs = {
                "name": var_name,
                "type_value": type_value,
                "base_value": str(var_settings.get("base", "")),
                "from_value": str(var_settings.get("from", "")),
                "to_value": str(var_settings.get("to", "")),
                "from_2_pow_value": str(var_settings.get("from_2^", "")),
                "to_2_pow_value": str(var_settings.get("to_2^", "")),
                "step_value": str(var_settings.get("step", "")),
                "from_type_value": str(var_settings.get("from", "")),
                "to_type_value": str(var_settings.get("to", "")),
                "op_value": str(var_settings.get("op", "")),
                "list_value": ", ".join(map(str, var_settings.get("list", []))),
                "source_value": str(var_settings.get("source", "")),
                "sources_value": ", ".join(map(str, var_settings.get("sources", []))),
                "format_value": str(var_keys.get("format", "")),
            }
            cards.append(variable_card(**card_kwargs))
        
        # Append Add card
        cards.append(add_card())
        return cards, generator_name, generator_template, gen_settings

    if cards is None:
        cards = []

    # Remove the Add card if present
    if cards and isinstance(cards[-1], dict) and cards[-1].get('props', {}).get('id') == "add-config-card":
        cards = cards[:-1]

    # Add new variable
    if trigger_id == "new-variable" and new_click:
        existing_names = [card.get('props', {}).get('id', {}).get('name', '') for card in cards if isinstance(card.get('props', {}).get('id', {}), dict)]
        var_idx = 1
        while f"var{var_idx}" in existing_names:
            var_idx += 1
        new_name = f"var{var_idx}"
        cards.append(variable_card(new_name))

    if isinstance(trigger_id, dict):
        trig_type = trigger_id.get("type")
        trig_name = trigger_id.get("name")

        # Delete
        if trig_type == "delete-var":
            cards = [
                card for card in cards
                if not (
                    isinstance(card.get('props', {}).get('id', {}), dict)
                    and card.get('props', {}).get('id', {}).get("type") == "variable-card"
                    and card.get('props', {}).get('id', {}).get("name") == trig_name
                )
            ]
        # Duplicate
        else:
            idx = None
            for i, card in enumerate(cards):
                card_id = card.get('props', {}).get('id', {})
                if isinstance(card_id, dict) and card_id.get("type") == "variable-card" and card_id.get("name") == trig_name:
                    idx = i
                    break

            if trig_type == "duplicate-var":

                # Find a unique name for the duplicated variable
                existing_names = [card.get('props', {}).get('id', {}).get('name', '') for card in cards if isinstance(card.get('props', {}).get('id', {}), dict)]
                copy_idx = 1
                while f"{trig_name}_copy{copy_idx}" in existing_names:
                    copy_idx += 1
                new_name = f"{trig_name}_copy{copy_idx}"

                if idx is not None:
                    cards.append(variable_card(
                        name=new_name,
                        type_value=types[idx],
                        base_value=base_vals[idx],
                        from_value=from_vals[idx],
                        to_value=to_vals[idx],
                        from_2_pow_value=from_2_pow_vals[idx],
                        to_2_pow_value=to_2_pow_vals[idx],
                        from_type_value=from_vals[idx],
                        to_type_value=to_vals[idx],
                        step_value=step_vals[idx],
                        op_value=op_vals[idx],
                        list_value=list_vals[idx],
                        source_value=sources_vals[idx],
                        sources_value=sources_vals[idx],
                    ))

    # Append Add card
    cards.append(add_card())
    return cards, dash.no_update, dash.no_update, dash.no_update

@dash.callback(
    Output({"type": "config-cards-row"}, "children"),
    Output("variable-preview", "children"),
    Output("gen-preview", "children"),
    Input("url", "search"),
    Input("url", "pathname"),
    Input({"page": page_path, "action": "save-all"}, "n_clicks"),
    Input({"action": "generate-all"}, "n_clicks"),
    Input("generator-name", "value"),
    Input("generator-template", "value"),
    Input({"type": "variable-title", "name": dash.ALL}, "value"),
    Input({"type": "variable-type", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-base", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-from", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-to", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-from_2_pow", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-to_2_pow", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-from_type", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-to_type", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-step", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-op", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-list", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-source", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-sources", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-format", "name": dash.ALL}, "value"),
    State({"type": "variable-metadata", "name": dash.ALL}, "data"),
    State("odatix-settings", "data"),
)
def update_generation(
    search, page, n_click_save, n_click_gen,
    name, template,
    titles, types, base_vals, from_vals, to_vals, from_2_pow_vals, to_2_pow_vals, from_type_vals, to_type_vals, step_vals, op_vals, list_vals, source_vals, sources_vals, format_vals,
    metadata, odatix_settings
):  
    trigger_id = ctx.triggered_id

    if trigger_id == "url":
        if page != page_path:
            return dash.no_update, dash.no_update, dash.no_update
    
    nb_vars = len(titles)
    if not (name and template and nb_vars > 0):
        return [], [], "0 configurations to be generated"
    
    gen_settings = get_gen_settings(
        name, template,
        titles, types, base_vals, from_vals, to_vals, from_2_pow_vals, to_2_pow_vals, from_type_vals, to_type_vals, step_vals, op_vals, list_vals, source_vals, sources_vals, format_vals
    )

    generator = ConfigGenerator(data=gen_settings)
    generated_params, variables = generator.generate()
    
    arch_path = odatix_settings.get("arch_path", OdatixSettings.DEFAULT_ARCH_PATH)
    arch_name = get_key_from_url(search, "arch")
    domain = get_key_from_url(search, "domain")
    if not domain:
        domain = hard_settings.main_parameter_domain
    if trigger_id == {"page": page_path, "action": "save-all"} or trigger_id == {"action": "generate-all"}:
        if domain and arch_name:
            workspace.update_domain_settings(
                arch_path=arch_path,
                arch_name=arch_name, 
                domain=domain, 
                settings_to_update=gen_settings,
            )
            if trigger_id == {"page": page_path, "action": "save-all"}:
                return dash.no_update, dash.no_update, dash.no_update
    
    if trigger_id == {"action": "generate-all"}:
        for config_name, config_content in generated_params.items():
            arch_domain_path = workspace.get_arch_domain_path(arch_path, arch_name, domain)
            config_file_path = os.path.join(arch_domain_path, f"{config_name}.txt")
            try:
                with open(config_file_path, "w") as config_file:
                    config_file.write(config_content)
            except Exception as e:
                pass
        return dash.no_update, dash.no_update, dash.no_update
        
    # Limit the number of previewed configurations
    if len(generated_params) > hard_settings.max_preview_values:
        preview_params = dict(list(generated_params.items())[:hard_settings.max_preview_values-2])
        ellipsis = True
        hidden_count = len(generated_params) - len(preview_params) - 1
    else:
        preview_params = generated_params
        ellipsis = False
        hidden_count = 0

    # Create config cards
    config_cards = []
    for config_name, config_content in preview_params.items():
        config_cards.append(config_card(
            filename=f"{config_name}.txt",
            content=config_content,
        ))
    
    if ellipsis:
        config_cards.append(config_card(
            filename="",
            content=f"{hidden_count} more configurations",
            ellipsis=True,
        ))
        config_name, config_content = list(generated_params.items())[-1]
        config_cards.append(config_card(
            filename=f"{config_name}.txt",
            content=config_content,
        ))

    variables_preview = []

    for var, values in variables.items():
        if len(values) > hard_settings.max_preview_values:
            values = list(values)[:hard_settings.max_preview_values-2] + ["[...]"] + [list(values)[-1]]
        variables_preview.append(
            html.Div(
                children=[
                    html.B(var),
                    html.Span(f": "),
                    html.Pre(
                        children=f"{', '.join(map(str, values))}",
                        style={"display": "inline", "paddingLeft": "5px", "paddingRight": "5px", "whiteSpace": "pre-wrap", "wordBreak": "break-word"},
                    ),
                ],
                style={"marginBottom": "5px"},
            )
        )
    
    title = f"{len(generated_params)} configurations to be generated"

    return config_cards, variables_preview, title

@dash.callback(
    Output("main-title-config-gen", "children"),
    Input("url", "search"),
    State("url", "pathname"),
    preview_initial_call=True
)
def update_main_domain_title(search, page):
    if page != page_path:
        return dash.no_update, dash.no_update, dash.no_update
    arch_name = get_key_from_url(search, "arch")
    domain = get_key_from_url(search, "domain")
    if not domain:
        domain = hard_settings.main_parameter_domain
        if not domain: domain = hard_settings.main_parameter_domain

    if not arch_name:
        return "No architecture selected."
    
    if domain == hard_settings.main_parameter_domain:
        title = f"{arch_name} - Main parameter domain"
    else:
        title = f"{arch_name} - {domain}"
    return title

@dash.callback(
    Output({"type": "more-variable-field-div", "name": dash.ALL}, "style"),
    Output({"type": "more-fields-icon", "name": dash.ALL}, "className"),
    Input({"type": "more-fields", "name": dash.ALL}, "n_clicks"),
    State({"type": "more-variable-field-div", "name": dash.ALL}, "style"),
    State({"type": "more-fields-icon", "name": dash.ALL}, "className"),
    State({"type": "variable-metadata", "name": dash.ALL}, "data"),
)
def toggle_more_fields(n_clicks, expandable_area_styles, icon_classes, metadata):
    # Get the button that triggered the callback
    trigger_id = ctx.triggered_id
    if not isinstance(trigger_id, dict) or "name" not in trigger_id:
        return [dash.no_update] * len(n_clicks), [dash.no_update] * len(n_clicks)

    # Update only the relevant variable card
    index = None
    for i, clicks in enumerate(n_clicks):
        current_name = metadata[i].get("name") if metadata and i < len(metadata) else {}
        if trigger_id.get("name") == current_name:
            index = i
            break
    
    new_expandable_area_styles = list(expandable_area_styles)
    new_icon_classes = list(icon_classes)
    if index is not None:
        if n_clicks[index] % 2 == 0:
            new_expandable_area_styles[index] = Style.hidden
            new_icon_classes[index] = "icon normal rotate"
        else:
            new_expandable_area_styles[index] = Style.visible
            new_icon_classes[index] = "icon normal rotate rotated"
    return new_expandable_area_styles, new_icon_classes

@dash.callback(
    Output({"page": page_path, "action": "save-all"}, "className"),
    Output("generator-saved-settings", "data"),
    Input({"page": page_path, "action": "save-all"}, "n_clicks"),
    Input({"action": "generate-all"}, "n_clicks"),
    Input("generator-name", "value"),
    Input("generator-template", "value"),
    Input({"type": "variable-title", "name": dash.ALL}, "value"),
    Input({"type": "variable-type", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-base", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-from", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-to", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-from_2_pow", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-to_2_pow", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-from_type", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-to_type", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-step", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-op", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-list", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-source", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-sources", "name": dash.ALL}, "value"),
    Input({"type": "variable-field-format", "name": dash.ALL}, "value"),
    State("generator-initial-settings", "data"),
    State("generator-saved-settings", "data"),
    State({"type": "variable-metadata", "name": dash.ALL}, "data"),
    prevent_initial_call=True
)
def update_save_button(
    save_n_clicks, generate_n_clicks, name, template, 
    title_values, type_values, base_vals, from_vals, to_vals, from_2_pow_vals, to_2_pow_vals, from_type_vals, to_type_vals, step_vals, op_vals, list_vals, source_vals, sources_vals, format_vals,
    initial_settings, saved_settings, metadata
):
    button_disabled = "color-button disabled icon-button"
    button_enabled = "color-button warning icon-button tooltip delay bottom small"

    trigger_id = ctx.triggered_id

    if trigger_id == {"page": page_path, "action": "save-all"} or trigger_id == {"action": "generate-all"}:
        return button_disabled, {
            "name": name,
            "template": template,
        }

    if saved_settings is None:
        settings = initial_settings
    else:
        settings = saved_settings

    old_name = settings.get("name", "") if settings else ""
    old_template = settings.get("template", "") if settings else ""

    if name != old_name or template != old_template:
        return button_enabled, dash.no_update
    
    for i, _ in enumerate(title_values):
        if str(title_values[i]) != str(metadata[i].get("name")):
            return button_enabled, dash.no_update
        if str(type_values[i]) != str(metadata[i].get("type")):
            return button_enabled, dash.no_update
        if str(base_vals[i]) != str(metadata[i].get("base_value")):
            return button_enabled, dash.no_update
        if str(from_vals[i]) != str(metadata[i].get("from_value")):
            return button_enabled, dash.no_update
        if str(to_vals[i]) != str(metadata[i].get("to_value")):
            return button_enabled, dash.no_update
        if str(from_2_pow_vals[i]) != str(metadata[i].get("from_2_pow_value")):
            return button_enabled, dash.no_update
        if str(to_2_pow_vals[i]) != str(metadata[i].get("to_2_pow_value")):
            return button_enabled, dash.no_update
        if str(from_type_vals[i]) != str(metadata[i].get("from_type_value")):
            return button_enabled, dash.no_update
        if str(to_type_vals[i]) != str(metadata[i].get("to_type_value")):
            return button_enabled, dash.no_update
        if str(step_vals[i]) != str(metadata[i].get("step_value")):
            return button_enabled, dash.no_update
        if str(op_vals[i]) != str(metadata[i].get("op_value")):
            return button_enabled, dash.no_update
        if str(list_vals[i]) != str(metadata[i].get("list_value")):
            return button_enabled, dash.no_update
        if str(source_vals[i]) != str(metadata[i].get("source_value")):
            return button_enabled, dash.no_update
        if str(sources_vals[i]) != str(metadata[i].get("sources_value")):
            return button_enabled, dash.no_update
        if str(format_vals[i]) != str(metadata[i].get("format_value")):
            return button_enabled, dash.no_update

    return button_disabled, dash.no_update

@dash.callback(
    Output(f"{page_path}-dummy1", "data"), # For older versions of Dash that do not support no_output
    Input({"action": "clean-all"}, "n_clicks"),
    State("url", "search"),
    State("odatix-settings", "data"),
)
def clean_all_configs(n_clicks, search, odatix_settings):    
    arch_path = odatix_settings.get("arch_path", OdatixSettings.DEFAULT_ARCH_PATH)
    arch_name = get_key_from_url(search, "arch")
    domain = get_key_from_url(search, "domain")
    if not domain:
        domain = hard_settings.main_parameter_domain
    
    trigger_id = ctx.triggered_id
    if trigger_id == {"action": "clean-all"} and n_clicks:
        workspace.delete_all_config_files(arch_path, arch_name, domain)
    
    return dash.no_update

@dash.callback(
    Output("config-gen-back-button", "href"),
    Input("url", "search"),
)
def update_back_button_link(search):
    arch_name = get_key_from_url(search, "arch")
    return f"/config_editor?arch={arch_name}"


######################################
# Layout
######################################

variable_title_tile_buttons = html.Div(
    children=[
        ui.save_button(
            id={"page": page_path, "action": "save-all"},
            tooltip="Save all changes",
            disabled=True,
        ),
    ],
    className="inline-flex-buttons",
)
preview_title_tile_buttons = html.Div(
    children=[
        ui.icon_button(
            icon=icon("clean", className="icon red"),
            color="caution",
            id={"action": "clean-all"},
            tooltip="Delete all existing configuration files in this parameter domain",
        ),
        ui.icon_button(
            icon=icon("generate", className="icon blue"),
            color="secondary",
            text="Generate", 
            id={"action": "generate-all"},
            tooltip="Generate all previewed configurations",
        ),
    ],
    className="inline-flex-buttons",
)


layout = html.Div(
    children=[
        dcc.Location(id="url"),
        html.Div(style={"marginTop": "10px"}),
        ui.title_tile(id="main-title-config-gen", buttons=variable_title_tile_buttons, back_button_link="/config_editor", back_button_id="config-gen-back-button"),
        html.Div(
            children=[
                config_parameters_form({}),
                html.Div(
                    children=[
                        html.H3(f"Variable Preview"),
                        html.Div(id="variable-preview")
                    ],
                    id="preview-pane",
                    className="tile config"
                ),
            ], 
            className="card-matrix config",
        ),
        ui.title_tile(text="Variable Definition", id="variable-title"),
        html.Div([ 
            html.Div(
                children=[
                    variable_card(name="var", type_value="range", from_value="1", to_value="10", step_value="1"),
                    add_card(),
                ],
                id={"type": "variable-cards-row"},
                className=f"card-matrix configs", 
                style={"marginLeft": "13px"},
            ),
        ]),
        ui.title_tile(text="Generation Preview", id="gen-preview", buttons=preview_title_tile_buttons),
        html.Div([ 
            html.Div(
                id={"type": "config-cards-row"},
                className=f"card-matrix configs", 
                style={"marginLeft": "13px", "marginBottom": "30px"},
            ),
        ]),
        dcc.Store(id={"type": "update_url", "id": page_path}),
        dcc.Store(id=f"{page_path}-dummy1"),
    ],
    className="page-content",
    style={
        "padding": "0 16%",
        "display": "flex",  
        "flexDirection": "column",
        "min-height": f"calc(100vh - {navigation.top_bar_height})",
    }
)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

from typing import Optional, Union
from dash_svg import Svg
from dash import html, dcc
from dash.development.base_component import Component

from odatix.gui.icons import icon

def icon_button(icon, color, text="", id=None, link=None, multiline=False, width="115px", style={}, tooltip:str="", tooltip_options:str="bottom"):  
    """
    Create a button with an icon and optional text.
    Args:
        icon (str or Svg): The icon to display (filename in assets/icons or Svg component).
        color (str): The color of the button (e.g., "red", "blue", "green").
        text (str, optional): The text to display next to the icon. Defaults to "" (just the icon).
        id (str, optional): The id of the button. Defaults to None.
        link (str | dict, optional): If provided, the button will be a link to this URL. Defaults to None.
        multiline (bool, optional): If True, adjust vertical alignment for multiline text. Defaults to False.
    """ 
    if text:
        style = {**style, "min-width": width}
    content = html.Button(
        html.Div(
            children=[
                icon if isinstance(icon, Svg) else html.Img(
                    src=f"/assets/icons/{icon}",
                    alt=text if text else icon.capitalize(),
                    style={
                        "width": "25px",
                        "height": "25px",
                        "marginLeft": "-10px",
                    }
                ),
                html.Span(
                    text,
                    style={"fontWeight": "600", "fontSize": "1em", "marginLeft": "5px"} if text else {}
                ),
            ],
            style={"display": "flex", "alignItems": "center", "justifyContent": "flex-start", "marginTop": "-9px" if multiline else "-4px", "width": "100%"},
        ),
        id=id if id else "",
        n_clicks=0,
        className=f"color-button {color} icon-button {f' tooltip {tooltip_options}' if tooltip else ''}",
        style=style,
        **{'data-tooltip': tooltip},
    )
    if link is None:
        return content
    else:
        link_kwargs = {
            "children": content,
            "href": link,
            "style": {"textDecoration": "none"}
        }
        if isinstance(id, dict):
            link_id = id.copy()
            link_id.update({"is_link": True})
        elif isinstance(id, str):
            link_id = id + "-link"
        else:   
            link_id = None
        if link_id is not None:
            link_kwargs["id"] = link_id
        return dcc.Link(**link_kwargs)

def delete_button(id, large=False, tooltip:str="Delete"):
    return icon_button(
        icon=icon("delete", width="25px", height="25px", className="icon red"),
        color="caution", 
        text="Delete" if large else "", 
        tooltip=tooltip,
        tooltip_options="bottom auto",
        id=id,
    )

def duplicate_button(id, large=False, tooltip:str="Duplicate"):
    return icon_button(
        icon=icon("duplicate", className="icon blue"),
        color="secondary",
        text="Duplicate" if large else "", 
        tooltip=tooltip,
        tooltip_options="bottom auto",
        id=id,
    )

def save_button(id, text="Save All", disabled=False, tooltip:str="Save all changes"):
    if isinstance(id, dict):
        icon_id = id.copy()
        icon_id.update({"is_icon": True})
    else:
        icon_id = id + "-icon"

    return icon_button(
        icon=icon("save", className="icon", id=icon_id),
        color="disabled" if disabled else "warning",
        text=text, 
        tooltip=tooltip,
        tooltip_options="bottom auto caution",
        id=id,
    )


def title_tile(text:str="", id:Union[str, dict]="main-title", buttons:html.Div=html.Div(), back_button_link:Optional[str]=None, back_button_id:Optional[str]=None, tooltip:str="", switch:Optional[bool]=None, style={}):
    if isinstance(id, dict):
        checklist_id = id.copy()
        checklist_id.update({"is_switch": True})
    else:
        checklist_id = id + "-switch"
    title_content = html.Div(
        children=[
            html.Div(
                children=[
                    dcc.Checklist(
                        options=[{"label": "", "value": True}],
                        value=[True] if switch else [],
                        id=checklist_id,
                        className="checklist-switch",
                        style={"display": "inline-block", "transform": "translate(-7px, -17px)"},
                    ) if switch is not None else html.Div(),
                    html.H3(text, id=id, style={"marginBottom": "0px", "display": "inline-block"}),
                    tooltip_icon(tooltip) if tooltip else html.Div(style={"display": "none"}),
                ],
            ),
            html.Div(
                [buttons],
            ),
        ],
        className="title-tile-flex",
        style={
            "display": "flex",
            "alignItems": "center",
            "padding": "0px",
            "justifyContent": "space-between",
        }
    )
    style = {**style, "position": "relative"}
    return html.Div(
        html.Div(
            children=[
                back_button(link=back_button_link, id=back_button_id),
                title_content
            ],
            className="tile title",
            style=style,
        ),
        className="card-matrix config",
        style={"marginTop": "10px", "marginBottom": "0px"},
    )

def subtitle_div(text:str="", id:str="main-title", buttons:html.Div=html.Div()):
    title_content = html.Div([
        html.H3(text, id=id, style={"marginBottom": "0px"}),
        html.Div(
            [buttons],
        ),
    ],
    className="title-tile-flex",
    style={
        "display": "flex",
        "alignItems": "center",
        "padding": "0px",
        "margin-right": "-15px",
        "margin-top": "0px",
        "justifyContent": "space-between",
    })
    return title_content

def back_button(link: Optional[str]="/", id: Optional[str]="back-button"):
    if link is None and id is None:
        return html.Div()
    if link:
        button = dcc.Link(
            icon("back", className="icon back-button", width="30px", height="30px"),
            href=link,
            id=id,
            style={"textDecoration": "none"},
        )
    else:
        button = html.Button(
            icon("back", className="icon back-button", width="30px", height="30px"),
            id=id,
            n_clicks=0,
            style={"background": "none", "border": "none", "padding": "0px", "margin": "0px"},
        )
    return html.Div(
        button,
        style={
            "position": "absolute",
            "display": "flex",
            "alignItems": "center",
            "left": "0",
            "top": "0",
            "height": "100%",
            "width": "45px",
            "transform": "translate(-45px)",
        },
    )

def tooltip_icon(tooltip: str="", tooltip_options: str="secondary") -> Component:
    return html.Div(
        children=[
            icon(
                "tooltip",
                className="icon",
                color="default",
                width="20px",
                height="20px",
            ),
        ],
        className="tooltip " + tooltip_options,
        style={"display": "inline-block", "transform": "translate(15px, 2px)", "verticalAlign": "middle"},
        **{"data-tooltip": tooltip},
    )

def create_card_button(page: dict) -> Component:
    options = {
        "className": "card home hover",
        "style": {"textDecoration": "none"},
        "children": [
            html.Img(
                src=page["image"],
                className="card-img",
                style={"height": "125px"}
            ),
            html.Div(
                page["name"],
                className="card-title",
            ),
            html.Div(
                page["description"],
                className="card-description",
            ),
        ],
    }
    if "id" in page:
        options["id"] = page["id"]
    if "link" in page:
        options["href"] = page["link"]
        if page["link"].startswith("/"):
            return dcc.Link(
                **options
            )
        else:
            options["target"] = "_blank"
            return html.A(
                **options
            )
    else:
        return html.Div(
            **options
        )



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import re
import yaml

from odatix.lib.get_from_dict import get_from_dict, Key, KeyNotInDictError, BadValueInDictError
import odatix.lib.printc as printc
import odatix.lib.hard_settings as hard_settings
from odatix.lib.settings import OdatixSettings

script_name = os.path.basename(__file__)

class ParamDomain:
  """
  Represents a parameter domain, which defines configuration settings
  for a given hardware architecture.

  Attributes:
      domain (str): Name of the parameter domain (e.g., architecture name).
      domain_value (str): Specific configuration variant.
      use_parameters (bool): Whether parameters are used in this domain.
      start_delimiter (str): Start delimiter for parameter extraction.
      stop_delimiter (str): Stop delimiter for parameter extraction.
      param_target_file (str): File where parameters should be applied.
      param_file (str): Path to the parameter definition file.
  """

  def __init__(
    self,
    domain,
    domain_value,
    use_parameters,
    start_delimiter,
    stop_delimiter,
    param_target_file,
    param_file,
  ):
    """Initializes a ParamDomain object with its attributes."""
    self.domain = domain
    self.domain_value = domain_value
    self.use_parameters = use_parameters
    self.start_delimiter = start_delimiter
    self.stop_delimiter = stop_delimiter
    self.param_target_file = param_target_file
    self.param_file = param_file
  
  def __repr__(self):
    """Returns a string representation of the object for debugging purposes."""
    out = f"domain: {self.domain}\n"
    out += f"domain_value: {self.domain_value}\n"
    out += f"use_parameters: {self.use_parameters}\n"
    out += f"start_delimiter: {self.start_delimiter}\n"
    out += f"stop_delimiter: {self.stop_delimiter}\n"
    out += f"param_target_file: {self.param_target_file}\n"
    out += f"param_file: {self.param_file}\n"
    return out

  @staticmethod
  def check_parameter_file(parameter_file, arch_param_dir, generate_enabled=False):
    """
    Checks if the parameter file exists.

    Args:
        parameter_file (str): Path to the parameter file.
        arch_param_dir (str): Directory of the architecture parameters.
        generate_enabled (bool): Whether configuration generation generation is enabled.

    Returns:
        bool: True if the file exists, False otherwise.
    """
    if not os.path.isfile(parameter_file):
      printc.error("The parameter file \"" + parameter_file + "\" does not exist in directory \"" + arch_param_dir + "\"", script_name)
      if generate_enabled:
        printc.note("Since \"generate_configutations\" is enabled:", script_name)
        printc.note("Did you run \"odatix generate\"?", script_name)
        printc.note("Are your generation settings correct?", script_name)
      return False
    return True

  @staticmethod
  def check_settings_file(settings_file, arch_param_dir):
    """
    Checks if the settings file exists.

    Args:
        settings_file (str): Path to the settings file.
        arch_param_dir (str): Directory of the architecture parameters.

    Returns:
        bool: True if the file exists, False otherwise.
    """
    if not os.path.isfile(settings_file):
      printc.error("The settings file \"" + settings_file + "\" does not exist in directory \"" + arch_param_dir + "\"", script_name)
      return False
    return True

  @staticmethod
  def get_param_domains(requested_param_domains, architecture, arch_path=OdatixSettings.DEFAULT_ARCH_PATH, param_settings_filename=hard_settings.param_settings_filename, top_level_file=""):
    """
    Retrieves multiple parameter domains.

    Args:
        requested_param_domains (list): List of requested parameter domains.
        architecture (str): Name of the architecture.
        arch_path (str): Path to the architecture directory.
        param_settings_filename (str): Name of the settings file.
        top_level_file (str): Path to the top-level configuration file.

    Returns:
        list[ParamDomain] or None: List of ParamDomain objects if successful, otherwise None.
    """
    if not isinstance(requested_param_domains, list):
      printc.error("requested_param_domains must be a list", script_name)
      return None

    param_domains = []
    for request in requested_param_domains:
      param_domain = ParamDomain.get_param_domain(
        request=request, 
        architecture=architecture,
        arch_path=arch_path, 
        param_settings_filename=param_settings_filename,
        top_level_file=top_level_file,
        param_domain_def=True,
      )
      if param_domain is None:
        return None
      param_domains.append(param_domain)
    return param_domains

  @staticmethod
  def get_param_domain(request, architecture, arch_path=OdatixSettings.DEFAULT_ARCH_PATH, param_settings_filename=hard_settings.param_settings_filename, top_level_file="", param_domain_def=False):
    """
    Retrieves a single parameter domain.

    Args:
        request (str): Requested parameter domain.
        architecture (str): Name of the architecture.
        arch_path (str): Path to the architecture directory.
        param_settings_filename (str): Name of the settings file.
        top_level_file (str): Path to the top-level configuration file.
        param_domain_def (bool): Whether this is the definition of a parameter domain.

    Returns:
        ParamDomain or None: A ParamDomain object if successful, otherwise None.
    """

    arch_path = os.path.join(arch_path, architecture)

    # get param dir (request name before '/')
    arch_param_dir = re.sub('/.*', '', request)

    # get configuration (request name after '/')
    arch_config = re.sub('.*/', '', request)

    param_domain_path = os.path.join(arch_path, arch_param_dir)

    parameter_file = os.path.join(param_domain_path, arch_config + '.txt')

    settings_file = os.path.join(param_domain_path, param_settings_filename)
    success = ParamDomain.check_settings_file(settings_file, param_domain_path)
    if not success:
      return None

    # Load settings file
    with open(settings_file, 'r') as f: 
      try:
        settings_data = yaml.load(f, Loader=yaml.loader.SafeLoader)
      except Exception as e:
        printc.error("Settings file \"" + settings_file + "\" is not a valid YAML file", script_name)
        printc.cyan("error details: ", end="", script_name=script_name)
        print(str(e))
        return None

    generate_enabled, _ = get_from_dict("generate_configutations", settings_data, settings_file, default_value=False, silent=True, script_name=script_name)
    success = ParamDomain.check_parameter_file(parameter_file, param_domain_path, generate_enabled)
    if not success:
      return None

    # Retrieve parameter delimiters and usage details
    use_parameters, start_delimiter, stop_delimiter, param_target_filename = ParamDomain.get_param_delimiters(
        settings_data, settings_file, top_level_file, param_domain_def
    )

    return ParamDomain(
      domain=arch_param_dir,
      domain_value=arch_config,
      use_parameters=use_parameters,
      start_delimiter=start_delimiter,
      stop_delimiter=stop_delimiter,
      param_target_file=param_target_filename,
      param_file=parameter_file,
    )

  @staticmethod
  def get_param_delimiters(settings_data, settings_file, top_level_file=None, param_domain_def=False):
    """
    Extracts parameter delimiters and target file information.

    Args:
        settings_data (dict): Parsed YAML settings data.
        settings_file (str): Path to the settings file.
        top_level_file (str): Path to the top-level configuration file.
        param_domain_def (bool): Whether this is a parameter domain definition.

    Returns:
        tuple: (use_parameters, start_delimiter, stop_delimiter, param_target_filename)
    """
    default_value = param_domain_def
    silent = param_domain_def

    use_parameters, _ = get_from_dict('use_parameters', settings_data, settings_file, default_value=default_value, silent=silent, type=bool, script_name=script_name)
    
    if use_parameters:
      start_delimiter, defined = get_from_dict('start_delimiter', settings_data, settings_file, silent=True, script_name=script_name)
      if not defined:
        printc.error("Cannot find key \"start_delimiter\" in \"" + settings_file + "\", while \"use_parameters\" is true", script_name)
        return None, None, None, None
      stop_delimiter, defined = get_from_dict('stop_delimiter', settings_data, settings_file, silent=True, script_name=script_name)
      if not defined:
        printc.error("Cannot find key \"stop_delimiter\" in \"" + settings_file + "\", while \"use_parameters\" is true", script_name)
        return None, None, None, None

      # Print an error if the default value is not defined
      silent = False if top_level_file is None else True
      param_target_filename, _ = get_from_dict('param_target_file', settings_data, settings_file, default_value=top_level_file, silent=silent, script_name=script_name)
      if param_target_filename == "":
        param_target_filename = top_level_file
    else:
      start_delimiter = ""
      stop_delimiter = ""
      param_target_filename = ""

    return use_parameters, start_delimiter, stop_delimiter, param_target_filename



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import re
import sys
import yaml
import shutil
import argparse
import subprocess

from odatix.components.replace_params import replace_params
import odatix.lib.printc as printc
import odatix.lib.hard_settings as hard_settings
from odatix.lib.parallel_job_handler import ParallelJobHandler, ParallelJob
from odatix.lib.settings import OdatixSettings
from odatix.lib.architecture_handler import ArchitectureHandler, Architecture
from odatix.lib.read_tool_settings import read_tool_settings
from odatix.lib.utils import read_from_list, copytree, create_dir, ask_to_continue, KeyNotInListError, BadValueInListError, get_timestamp_string
from odatix.lib.get_from_dict import get_from_dict
from odatix.lib.prepare_work import edit_config_file
from odatix.lib.check_tool import check_tool
from odatix.lib.run_settings import get_synth_settings
from odatix.lib.variables import replace_variables, Variables

script_name = os.path.basename(__file__)


######################################
# Parse Arguments
######################################


def add_arguments(parser):
  parser.add_argument("-t", "--tool", default="vivado", help="eda tool in use (default: vivado)")
  parser.add_argument("-o", "--overwrite", action="store_true", help="overwrite existing results")
  parser.add_argument("-y", "--noask", action="store_true", help="do not ask to continue")
  parser.add_argument("-i", "--input", help="input settings file")
  parser.add_argument("-a", "--archpath", help="architecture directory")
  parser.add_argument("-w", "--work", help="work directory")
  parser.add_argument("-E", "--exit", action="store_true", help="exit monitor when all jobs are done")
  parser.add_argument("-j", "--jobs", help="maximum number of parallel jobs")
  parser.add_argument("-T", "--trust", action="store_true", help="do not check eda tool before runnning jobs (saves time)")
  parser.add_argument("-D", "--debug", action="store_true", help="enable debug mode to help troubleshoot settings files")
  parser.add_argument("--from", dest="from_freq", type=int, help="override range lower bound for custom frequency synthesis (in MHz)")
  parser.add_argument("--to", dest="to_freq", type=int, help="override range upper bound for custom frequency synthesis (in MHz)")
  parser.add_argument("--step", dest="step_freq", type=int, help="override range step bound for custom frequency synthesis (in MHz)")
  parser.add_argument("--at", dest="at_freq", action='append', type=int, help="override freqency at which custom frequency synthesis should be run (in MHz)")
  parser.add_argument("-k", "--keep", action="store_true", help="store synthesis batch with a timestamp in the configuration name")
  parser.add_argument("--logsize", help="size of the log history per job in the monitor")
  parser.add_argument(
    "-c",
    "--config",
    default=OdatixSettings.DEFAULT_SETTINGS_FILE,
    help="global settings file for Odatix (default: " + OdatixSettings.DEFAULT_SETTINGS_FILE + ")",
  )


def parse_arguments():
  parser = argparse.ArgumentParser(description="Run synthesis at specigied frequencies on selected architectures")
  add_arguments(parser)
  return parser.parse_args()


######################################
# Run Synthesis
######################################


def run_synthesis(run_config_settings_filename, arch_path, tool, work_path, target_path, overwrite, noask, exit_when_done, log_size_limit, nb_jobs, check_eda_tool, custom_freq_list=[], debug=False, keep=False):
  _overwrite, ask_continue, _exit_when_done, _log_size_limit, _nb_jobs, architectures = get_synth_settings(run_config_settings_filename)

  work_path = os.path.join(work_path, tool)

  if architectures is None:
    printc.error('The "architectures" section of "' + run_config_settings_filename + '" is empty.', script_name)
    printc.note('You must define your architectures in "' + run_config_settings_filename + '" before using this command.', script_name)
    printc.note("Check out examples Odatix's documentation for more information.", script_name)
    sys.exit(-1)

  if overwrite:
    overwrite = True
  else:
    overwrite = _overwrite

  if exit_when_done:
    exit_when_done = True
  else:
    exit_when_done = _exit_when_done

  if log_size_limit is not None:
    log_size_limit = int(log_size_limit)
  else:
    log_size_limit = _log_size_limit

  if nb_jobs is not None:
    nb_jobs = int(nb_jobs)
  else:
    nb_jobs = _nb_jobs

  if noask:
    ask_continue = False

  eda_target_filename = os.path.realpath(os.path.join(target_path, "target_" + tool + ".yml"))

  # Check if the target file exists
  if not os.path.isfile(eda_target_filename):
    printc.error(
      'Target file "' + eda_target_filename + '", for the selected eda tool "' + tool + '" does not exist', script_name
    )
    sys.exit(-1)

  # Check if the tool has a dedicated directory in odatix_eda_tools path
  eda_tool_dir = os.path.join(OdatixSettings.odatix_eda_tools_path, tool)
  if not os.path.isdir(eda_tool_dir):
    printc.error(
      'The directory "' + eda_tool_dir + '", for the selected eda tool "' + tool + '" does not exist', script_name
    )
    if tool not in hard_settings.default_supported_tools:
      printc.note(
        'The selected eda tool "'
        + tool
        + "\" is not one of the supported tool. Check out Odatix's documentation to add support for your own eda tool",
        script_name,
      )
    sys.exit(-1)

  # Get tool settings
  tool_settings_file = os.path.realpath(os.path.join(eda_tool_dir, hard_settings.tool_settings_filename))
  process_group, report_path, run_command, tool_test_command, _ = read_tool_settings(tool, tool_settings_file,  synth_type='custom_freq_synthesis')

  with open(eda_target_filename, "r") as f:
    try:
      settings_data = yaml.load(f, Loader=yaml.loader.SafeLoader)
    except Exception as e:
      printc.error('Settings file "' + eda_target_filename + '" is not a valid YAML file', script_name)
      printc.cyan("error details: ", end="", script_name=script_name)
      print(str(e))
      sys.exit(-1)

    # Mandatory keys
    try:
      targets = read_from_list("targets", settings_data, eda_target_filename, script_name=script_name)
      constraint_file = read_from_list("constraint_file", settings_data, eda_target_filename, script_name=script_name)
    except (KeyNotInListError, BadValueInListError):
      sys.exit(-1)  # if a key is missing

    # Optional keys
    try:
      install_path = read_from_list("tool_install_path", settings_data, eda_target_filename, print_error=False, script_name=script_name)
      install_path = os.path.realpath(os.path.expanduser(install_path))
      if not os.path.isdir(install_path):
        printc.error('The installation path "' + install_path + '" defined for tool "' + tool + '" in "' + eda_target_filename + '" does not exist', script_name)
        printc.note('Please update the path in "' + eda_target_filename + '"', script_name=script_name)
        printc.note('if no installation path is needed by ' + tool + '\'s Makefile, simply remove "install_path" from "' + eda_target_filename + '"', script_name=script_name)
        sys.exit(-1)

    except (KeyNotInListError, BadValueInListError):
      printc.note('No tool_install_path specified for "' + tool + '"', script_name=script_name)
      install_path = "/"

    force_single_thread, _ = get_from_dict("force_single_thread", settings_data, eda_target_filename, default_value=False, script_name=script_name)

  # Concat all strings if it is a list
  if isinstance(tool_test_command, list):
    tool_test_command = " ".join(map(str, tool_test_command)) 

  # Define user accessible variables
  variables = Variables(
    tool_install_path=os.path.realpath(install_path),
    odatix_path=OdatixSettings.odatix_path,
    odatix_eda_tools_path=OdatixSettings.odatix_eda_tools_path,
  )

  # Replace variables in command
  tool_test_command = replace_variables(tool_test_command, variables)

  # Try launching eda tool
  if check_eda_tool:
    check_tool(
      tool, command=tool_test_command, supported_tools=hard_settings.default_supported_tools, tool_install_path=install_path, debug=debug
    )

  ParallelJob.set_patterns(hard_settings.synth_status_pattern, hard_settings.fmax_status_pattern)

  arch_handler = ArchitectureHandler(
    work_path=work_path,
    arch_path=arch_path,
    script_path=OdatixSettings.odatix_eda_tools_path,
    log_path=hard_settings.work_log_path,
    work_rtl_path=hard_settings.work_rtl_path,
    work_script_path=hard_settings.work_script_path,
    work_report_path=hard_settings.work_report_path,
    work_log_path=hard_settings.work_log_path,
    process_group=process_group,
    command=run_command,
    eda_target_filename=eda_target_filename,
    fmax_status_filename=hard_settings.synth_status_filename,
    frequency_search_filename=hard_settings.frequency_search_filename,
    param_settings_filename=hard_settings.param_settings_filename,
    valid_status=hard_settings.valid_status,
    valid_frequency_search=hard_settings.valid_frequency_search,
    forced_fmax_lower_bound=None,
    forced_fmax_upper_bound=None,
    forced_custom_freq_list=custom_freq_list,
    overwrite=overwrite,
    force_single_thread=force_single_thread
  )

  timestamp = get_timestamp_string()
  architecture_instances = arch_handler.get_architectures(architectures, targets, constraint_file, install_path, range_mode=True,  keep=keep, timestamp=timestamp)

  # Print checklist summary
  arch_handler.print_summary()

  # ask to quit or continue
  valid_arch_count = arch_handler.get_valid_arch_count()
  if valid_arch_count > 0:
    if ask_continue:
      printc.bold("\nTotal: " + str(valid_arch_count))
      ask_to_continue()
  else:
    sys.exit(-1)

  print()

  job_list = []

  def prepare_job(arch_instance):
    if True:
      # Get param dir (arch name before '/')
      arch_param_dir = re.sub("/.*", "", arch_instance.arch_name)

      # Create directory
      create_dir(arch_instance.tmp_dir)

      # Create log dir
      create_dir(arch_instance.tmp_log_path)
      
      # Copy scripts
      try:
        copytree(os.path.join(OdatixSettings.odatix_eda_tools_path, hard_settings.common_script_path), arch_instance.tmp_script_path)
      except:
        printc.error('"' + arch_instance.tmp_script_path + '" exists while it should not', script_name)

      copytree(os.path.join(OdatixSettings.odatix_eda_tools_path, tool, hard_settings.tool_tcl_path), arch_instance.tmp_script_path, dirs_exist_ok=True)

      # Copy design
      if arch_instance.design_path is not None:
        copytree(
          src=arch_instance.design_path,
          dst=arch_instance.tmp_dir,
          whitelist=arch_instance.design_path_whitelist,
          blacklist=arch_instance.design_path_blacklist,
          dirs_exist_ok=True
        )

      # Copy rtl (if exists)
      if not arch_instance.generate_rtl:
        copytree(arch_instance.rtl_path, os.path.join(arch_instance.tmp_dir, hard_settings.work_rtl_path), dirs_exist_ok=True)

      # Replace parameters
      if arch_instance.use_parameters:
        if debug: 
          printc.subheader("Replace main parameters")
        param_target_file = os.path.join(arch_instance.tmp_dir, arch_instance.param_target_filename)
        param_filename = os.path.join(arch_path, arch_instance.arch_name + ".txt")
        replace_params(
          base_text_file=param_target_file,
          replacement_text_file=param_filename,
          output_file=param_target_file,
          start_delimiter=arch_instance.start_delimiter,
          stop_delimiter=arch_instance.stop_delimiter,
          replace_all_occurrences=False,
          silent=False if debug else True,
        )
        if debug: 
          print()

      # Replace domain parameters
      domain_dict=dict()
      nb_domain = 0
      arch_config = re.sub('.*/', '', arch_instance.arch_name)
      domain_dict["__main__"] = arch_config
      if timestamp is not None:
        domain_dict["__timestamp__"] = timestamp 
      for param_domain in arch_instance.param_domains:
        if param_domain.use_parameters:
          param_target_file = os.path.join(arch_instance.tmp_dir, param_domain.param_target_file)
          if debug: 
            printc.subheader("Replace parameters for \"" + param_domain.domain + "/" + param_domain.domain_value+ "\"")
          success = replace_params(
            base_text_file=param_target_file,
            replacement_text_file=param_domain.param_file,
            output_file=param_target_file,
            start_delimiter=param_domain.start_delimiter,
            stop_delimiter=param_domain.stop_delimiter,
            replace_all_occurrences=False,
            silent=False if debug else True,
          )
          if success:
            nb_domain = nb_domain + 1
            domain_dict[param_domain.domain] = param_domain.domain_value
          if debug: 
            print()

      with open(os.path.join(arch_instance.tmp_dir, hard_settings.param_domains_filename), 'w') as param_domains_file:
        yaml.dump(domain_dict, param_domains_file, default_flow_style=False, sort_keys=False)

      # Create target and architecture files
      f = open(os.path.join(arch_instance.tmp_dir, hard_settings.target_filename), "w")
      print(arch_instance.target, file=f)
      f.close()
      f = open(os.path.join(arch_instance.tmp_dir, hard_settings.arch_filename), "w")
      print(arch_instance.arch_name, file=f)
      f.close()

      # File copy
      if arch_instance.file_copy_enable:
        file_copy_dest = os.path.join(arch_instance.tmp_dir, arch_instance.file_copy_dest)
        try:
          shutil.copy2(arch_instance.file_copy_source, file_copy_dest)
        except Exception as e:
          printc.error(
            'Could not copy "' + arch_instance.script_copy_source + '" to "' + os.path.realpath(file_copy_dest) + '"',
            script_name,
          )
          printc.cyan("error details: ", end="", script_name=script_name)
          print(str(e))
          return

      # Script copy
      if arch_instance.script_copy_enable:
        try:
          shutil.copy2(arch_instance.script_copy_source, arch_instance.tmp_script_path)
        except Exception as e:
          printc.error(
            'Could not copy "'
            + arch_instance.script_copy_source
            + '" to "'
            + os.path.realpath(arch_instance.tmp_script_path)
            + '"',
            script_name,
          )
          printc.cyan("error details: ", end="", script_name=script_name)
          print(str(e))
          return

      # Edit tcl config script
      tcl_config_file = os.path.join(arch_instance.tmp_script_path, hard_settings.tcl_config_filename)
      report_path = os.path.join(arch_instance.tmp_dir, hard_settings.work_report_path)
      edit_config_file(arch_instance, tcl_config_file)

      # Write yaml config script
      yaml_config_file = os.path.join(arch_instance.tmp_dir, hard_settings.yaml_config_filename)
      Architecture.write_yaml(arch_instance, yaml_config_file)

      # Link all scripts to config script
      for filename in os.listdir(arch_instance.tmp_script_path):
        if filename.endswith(".tcl"):
          with open(os.path.join(arch_instance.tmp_script_path, filename), "r") as f:
            tcl_content = f.read()
          pattern = re.escape(hard_settings.source_tcl) + r"(.+?\.tcl)"

          def replace_path(match):
            return "source " + os.path.join(os.path.realpath(arch_instance.tmp_script_path), match.group(1)).replace('\\','/')

          tcl_content = re.sub(pattern, replace_path, tcl_content)
          with open(os.path.join(arch_instance.tmp_script_path, filename), "w") as f:
            f.write(tcl_content)

      # Concat all strings if it is a list
      if isinstance(arch_handler.command, list):
        command = " ".join(map(str, arch_handler.command)) 
      else:
        command = arch_handler.command

      # Define user accessible variables
      variables = Variables(
        work_path=os.path.realpath(arch_instance.tmp_dir),
        tool_install_path=os.path.realpath(arch_instance.install_path),
        odatix_path=OdatixSettings.odatix_path,
        odatix_eda_tools_path=OdatixSettings.odatix_eda_tools_path,
        script_path=os.path.realpath(os.path.join(arch_instance.tmp_dir, hard_settings.work_script_path)),
        log_path=os.path.realpath(os.path.join(arch_instance.tmp_dir, hard_settings.work_log_path)),
        clock_signal=arch_instance.clock_signal,
        top_level_module=arch_instance.top_level_module,
        lib_name=arch_instance.lib_name,
      )

      # Replace variables in command
      command = replace_variables(command, variables)

      fmax_status_file = os.path.join(arch_instance.tmp_dir, hard_settings.work_log_path, hard_settings.fmax_status_filename)
      synth_status_file = os.path.join(arch_instance.tmp_dir, hard_settings.work_log_path, hard_settings.synth_status_filename)

      # Run custom frequency synthesis script
      running_arch = ParallelJob(
        process=None,
        command=command,
        directory=".",
        generate_rtl=arch_instance.generate_rtl,
        generate_command=arch_instance.generate_command,
        target=arch_instance.target,
        arch=arch_instance.arch_name,
        display_name=arch_instance.arch_display_name,
        status_file=fmax_status_file,
        progress_file=synth_status_file,
        tmp_dir=arch_instance.tmp_dir,
        log_size_limit=log_size_limit,
        progress_mode="synth",
        status="idle",
      )

      job_list.append(running_arch)

  for arch_instance in architecture_instances:
    prepare_job(arch_instance)

  parallel_jobs = ParallelJobHandler(
    job_list,
    nb_jobs,
    arch_handler.process_group,
    auto_exit=exit_when_done,
    format_yaml= tool_settings_file,
    log_size_limit=log_size_limit,
  )
  job_exit_success = parallel_jobs.run()


######################################
# Main
######################################


def main(args, settings=None):
  # Get settings
  if settings is None:
    settings = OdatixSettings(args.config)
    if not settings.valid:
      sys.exit(-1)

  if args.input is not None:
    run_config_settings_filename = args.input
  else:
    run_config_settings_filename = settings.custom_freq_synthesis_settings_file

  if args.archpath is not None:
    arch_path = args.archpath
  else:
    arch_path = settings.arch_path

  if args.work is not None:
    work_path = args.work
  else:
    work_path = os.path.join(settings.work_path, settings.custom_freq_synthesis_work_path)

  target_path = settings.target_path
  tool = args.tool
  overwrite = args.overwrite
  noask = args.noask
  exit_when_done = args.exit
  log_size_limit = args.logsize
  nb_jobs = args.jobs
  check_eda_tool = not args.trust
  debug = args.debug
  keep = args.keep

  if args.at_freq is None:
    custom_freq_list = []
  else:
    custom_freq_list = args.at_freq

  if args.to_freq is not None and (args.from_freq is None or args.step_freq is None):
    printc.error("--to cannot be used without --from and --step", script_name=script_name)
    sys.exit(-1)
  elif args.from_freq is not None and (args.to_freq is None or args.step_freq is None):
    printc.error("--from cannot be used without --to and --step", script_name=script_name)
    sys.exit(-1)
  elif args.step_freq is not None and (args.to_freq is None or args.from_freq is None):
    printc.error("--step cannot be used without --from and --to", script_name=script_name)
    sys.exit(-1)
  elif args.from_freq is not None and args.to_freq is not None and args.step_freq is not None:
    if ArchitectureHandler.check_bounds(args.from_freq, args.to_freq, args.step_freq, synth_type="custom frequency synthesis"):
      range_list = ArchitectureHandler.create_list_from_range(args.from_freq, args.to_freq, args.step_freq)
      custom_freq_list = custom_freq_list + range_list
    else:
      sys.exit(-1)

  run_synthesis(run_config_settings_filename, arch_path, tool, work_path, target_path, overwrite, noask, exit_when_done, log_size_limit, nb_jobs, check_eda_tool, custom_freq_list, debug, keep)


if __name__ == "__main__":
  args = parse_arguments()
  main(args)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

"""
This module defines the settings and initialization functions for Odatix.

Classes:
    OdatixSettings: Encapsulates all paths, settings, and initialization logic for Odatix.

Functions:
    None defined at the module level; all logic is encapsulated in the OdatixSettings class.

Usage:
    Use `OdatixSettings` to handle configuration file paths, work directories, and
    initialization procedures for Odatix.

    Example:
        settings = OdatixSettings()
        if settings.valid:
            print("Settings loaded successfully")
"""

import os
import sys
import yaml
from datetime import datetime

import odatix.lib.printc as printc
import odatix.components.motd as motd
from odatix.lib.get_from_dict import get_from_dict, Key, KeyNotInDictError, BadValueInDictError
from odatix.lib.utils import ask_yes_no, YAML_BOOL, copytree

current_dir = os.path.dirname(os.path.abspath(__file__))
script_name = os.path.basename(__file__)

######################################
# Settings
######################################

DEFAULT_SETTINGS_FILE = "odatix.yml"

# Determine the install path of Odatix
if getattr(sys, "frozen", False):
    base_path = os.path.dirname(sys.executable)
else:
    base_path = current_dir

class OdatixSettings:
    """
    Handles Odatix configuration, paths, and initialization.

    Attributes:
        DEFAULT_* (str): Default paths and filenames for Odatix configurations.
        odatix_*_path (str): Paths to key Odatix components (e.g., EDA tools definition, embedded examples).
        valid (bool): Indicates whether the settings were successfully loaded.
        settings_file_exists (bool): Whether the settings file exists.

    Methods:
        __init__: Initializes the settings object and loads configurations.
        read_settings_file: Reads and validates the Odatix settings file.
        init_examples: Copies example configurations to the current directory.
        init_path: Initializes the required directory structure for Odatix.
        init_directory_dialog: Creates the configuration with user prompts.
        init_directory_nodialog: Creates the configuration without user prompts.
        init_success: Displays success messages post-initialization.
    """

    DEFAULT_SETTINGS_FILE = DEFAULT_SETTINGS_FILE

    DEFAULT_WORK_PATH = "work"
    DEFAULT_SIMULATION_WORK_PATH = "simulations"
    DEFAULT_FMAX_SYNTHESIS_WORK_PATH = "fmax_synthesis"
    DEFAULT_CUSTOM_FREQ_SYNTHESIS_WORK_PATH = "custom_freq_synthesis"
    DEFAULT_RESULT_PATH = "results"
    DEFAULT_USERCONFIG_PATH = "odatix_userconfig"
    DEFAULT_ARCH_PATH = os.path.join(DEFAULT_USERCONFIG_PATH, "architectures")
    DEFAULT_SIM_PATH = os.path.join(DEFAULT_USERCONFIG_PATH, "simulations")
    DEFAULT_TARGET_PATH = DEFAULT_USERCONFIG_PATH
    DEFAULT_USE_BENCHMARK = False
    DEFAULT_BENCHMARK_FILE = "results/benchmark.yml"
    DEFAULT_CLEAN_SETTINGS_FILE = os.path.join(DEFAULT_USERCONFIG_PATH, "clean.yml")
    DEFAULT_SIMULATION_SETTINGS_FILE = os.path.join(DEFAULT_USERCONFIG_PATH, "simulations_settings.yml")
    DEFAULT_FMAX_SYNTHESIS_SETTINGS_FILE = os.path.join(DEFAULT_USERCONFIG_PATH, "fmax_synthesis_settings.yml")
    DEFAULT_CUSTOM_FREQ_SYNTHESIS_SETTINGS_FILE = os.path.join(DEFAULT_USERCONFIG_PATH, "custom_freq_synthesis_settings.yml")
    
    odatix_path = os.path.realpath(os.path.join(base_path, os.pardir))
    odatix_eda_tools_path = os.path.realpath(os.path.join(odatix_path, os.pardir, "odatix_eda_tools"))
    odatix_init_path = os.path.realpath(os.path.join(odatix_path, os.pardir, "odatix_init"))
    odatix_examples_path = os.path.realpath(os.path.join(odatix_path, os.pardir, "odatix_examples"))

    def __init__(self, settings_filename=DEFAULT_SETTINGS_FILE, silent=False, verbose=False):
        """
        Initialize Odatix settings by reading the configuration file.

        Args:
            settings_filename (str): Name of the settings file to load.
        """
        self.verbose = verbose
        self.settings_file_exists = os.path.isfile(settings_filename)
        if self.settings_file_exists:
            success = self.read_settings_file(settings_filename)
        else:
            if not silent:
                printc.warning("This directory does not contain an Odatix settings file \"" + settings_filename + "\".", script_name=script_name)
                printc.note("Run Odatix with the '--init' flag to generate it alongside other necessary config files", script_name=script_name)
            self.valid = False

    @staticmethod
    def get_settings_file_dict(settings_filename=DEFAULT_SETTINGS_FILE, silent=False, verbose=False):
        """
        Static method to load Odatix settings from a file.

        Args:
            settings_filename (str): Name of the settings file to load.
            silent (bool): If True, suppresses warning messages.
            verbose (bool): If True, enables verbose output.
        Returns:
            dict | None: Dictionary of settings if successful, None otherwise.
        """
        if not os.path.isfile(settings_filename):
            printc.note("Odatix settings file \"" + settings_filename + "\" does not exist.", script_name=script_name)
            settings_data = {}
        else:
            with open(settings_filename, "r") as f:
                try:
                    settings_data = yaml.load(f, Loader=yaml.loader.SafeLoader)
                except Exception as e:
                    printc.error("Settings file \"" + settings_filename + "\" is not a valid YAML file", script_name)
                    printc.cyan("error details: ", end="", script_name=script_name)
                    print(str(e))
                    return None
        return settings_data

    @staticmethod
    def get_settings_template_dict(settings_filename=DEFAULT_SETTINGS_FILE, silent=False, verbose=False, settings=None):
        """
        Static method to get a template dictionary for Odatix settings.

        Returns:
            dict: Template dictionary with default settings.
        """
        if settings is None:
            settings = OdatixSettings.get_settings_file_dict(settings_filename, silent=silent, verbose=verbose)
            if settings is None:
                settings = {}
        settings_template = {
            # Main paths
            "arch_path": settings.get('arch_path', ""),
            "sim_path": settings.get('sim_path', ""),
            "target_path": settings.get('target_path', ""),

            # Settings files
            "clean_settings_file": settings.get('clean_settings_file', ""),
            "simulation_settings_file": settings.get('simulation_settings_file', ""),
            "fmax_synthesis_settings_file": settings.get('fmax_synthesis_settings_file', ""),
            "custom_freq_synthesis_settings_file": settings.get('custom_freq_synthesis_settings_file', ""),

            # Work directories
            "work_path": settings.get('work_path', ""),
            "simulation_work_path": settings.get('simulation_work_path', ""),
            "fmax_synthesis_work_path": settings.get('fmax_synthesis_work_path', ""),
            "custom_freq_synthesis_work_path": settings.get('custom_freq_synthesis_work_path', ""),
            
            # Results
            "result_path": settings.get('result_path', ""),
            "use_benchmark": settings.get('use_benchmark', ""),
            "benchmark_file": settings.get('benchmark_file', ""),
        }
        return settings_template

    def read_settings_file(self, settings_filename=DEFAULT_SETTINGS_FILE):
        """
        Read and parse the settings file for Odatix configurations.

        Args:
            settings_filename (str): Path to the settings file.

        Returns:
            bool: True if the settings were successfully loaded, False otherwise.
        """
        silent = not self.verbose

        settings_data = OdatixSettings.get_settings_file_dict(settings_filename, silent=silent)
        if settings_data is None:
            self.valid = False
            return False

        # Retrieve values from the settings file
        self.work_path, _ = get_from_dict("work_path", settings_data, settings_filename, default_value=OdatixSettings.DEFAULT_WORK_PATH, silent=silent, script_name=script_name)
        self.simulation_work_path, _ = get_from_dict("simulation_work_path", settings_data, settings_filename, default_value=OdatixSettings.DEFAULT_SIMULATION_WORK_PATH, silent=silent, script_name=script_name)
        self.fmax_synthesis_work_path, _ = get_from_dict("fmax_synthesis_work_path", settings_data, settings_filename, default_value=OdatixSettings.DEFAULT_FMAX_SYNTHESIS_WORK_PATH, silent=silent, script_name=script_name)
        self.custom_freq_synthesis_work_path, _ = get_from_dict("custom_freq_synthesis_work_path", settings_data, settings_filename, default_value=OdatixSettings.DEFAULT_CUSTOM_FREQ_SYNTHESIS_WORK_PATH, silent=silent, script_name=script_name)
        self.result_path, _ = get_from_dict("result_path", settings_data, settings_filename, default_value=OdatixSettings.DEFAULT_RESULT_PATH, silent=silent, script_name=script_name)
        self.arch_path, _ = get_from_dict("arch_path", settings_data, settings_filename, default_value=OdatixSettings.DEFAULT_ARCH_PATH, silent=silent, script_name=script_name)
        self.sim_path, _ = get_from_dict("sim_path", settings_data, settings_filename, default_value=OdatixSettings.DEFAULT_SIM_PATH, silent=silent, script_name=script_name)
        self.target_path, _ = get_from_dict("target_path", settings_data, settings_filename, default_value=OdatixSettings.DEFAULT_TARGET_PATH, silent=silent, script_name=script_name)
        self.use_benchmark, _ = get_from_dict("use_benchmark", settings_data, settings_filename, default_value=OdatixSettings.DEFAULT_USE_BENCHMARK, type=bool, silent=silent, script_name=script_name)
        self.benchmark_file, _ = get_from_dict("benchmark_file", settings_data, settings_filename, default_value=OdatixSettings.DEFAULT_BENCHMARK_FILE, silent=silent, script_name=script_name)
        self.clean_settings_file, _ = get_from_dict("clean_settings_file", settings_data, settings_filename, default_value=OdatixSettings.DEFAULT_CLEAN_SETTINGS_FILE, silent=silent, script_name=script_name)
        self.simulation_settings_file, _ = get_from_dict("simulation_settings_file", settings_data, settings_filename, default_value=OdatixSettings.DEFAULT_SIMULATION_SETTINGS_FILE, silent=silent, script_name=script_name)
        self.fmax_synthesis_settings_file, _ = get_from_dict("fmax_synthesis_settings_file", settings_data, settings_filename, default_value=OdatixSettings.DEFAULT_FMAX_SYNTHESIS_SETTINGS_FILE , silent=silent, script_name=script_name)
        self.custom_freq_synthesis_settings_file, _ = get_from_dict("custom_freq_synthesis_settings_file", settings_data, settings_filename, default_value=OdatixSettings.DEFAULT_CUSTOM_FREQ_SYNTHESIS_SETTINGS_FILE , silent=silent, script_name=script_name)
        
        # Depreciation warnings
        no_longer_supported = False
        if "sim_work_path" in settings_data:
            printc.warning("\"sim_work_path\" is no longer supported, use \"simulation_work_path\" instead", script_name=script_name)
            no_longer_supported = True
        if "fmax_work_path" in settings_data:
            printc.warning("\"fmax_work_path\" is no longer supported, use \"fmax_synthesis_work_path\" instead", script_name=script_name)
            no_longer_supported = True
        if "custom_freq_work_path" in settings_data:
            printc.warning("\"custom_freq_work_path\" is no longer supported, use \"custom_freq_synthesis_work_path\" instead", script_name=script_name)
            no_longer_supported = True
        if no_longer_supported:
            printc.note("\"simulation_work_path\", \"fmax_synthesis_work_path\" and \"custom_freq_synthesis_work_path\" are relative to \"work_path\"", script_name=script_name)
        
        self.result_types = {
            "custom_freq_synthesis": {
                "key": "custom_freq_synthesis",
                "path": self.custom_freq_synthesis_work_path,
            },
            "fmax_synthesis": {
                "key": "fmax_synthesis",
                "path": self.fmax_synthesis_work_path,
            },
        }

        self.valid = True
        return True
        
    @staticmethod
    def init_examples():
        """
        Copy example configurations to the current directory.

        Returns:
            bool: True if successful, False otherwise.
        """
        try:
            src_path = OdatixSettings.odatix_examples_path
            dst_path = os.getcwd()
            copytree(src_path, dst_path, dirs_exist_ok=True)
        except Exception as e:
            printc.error("Could not copy examples: " + str(e), script_name=script_name)
            return False
        return True

    @staticmethod
    def init_path():
        """
        Initialize the required directory structure for Odatix.

        Returns:
            bool: True if successful, False otherwise.
        """
        try:
            src_path = OdatixSettings.odatix_init_path
            dst_path = os.getcwd()
            copytree(src_path, dst_path, dirs_exist_ok=True)
        except Exception as e:
            printc.error("Could not initialize directory: " + str(e), script_name=script_name)
            return False
        return True

    @staticmethod
    def init_directory_dialog(include_examples=None, prog=""):
        """
        Initialize the configuration directory with user interaction.

        Args:
            include_examples (bool | None): Whether to include example files (prompted if None).
            prog (str): Program name for success message.

        Returns:
            bool: True if initialization is successful, False otherwise.
        """
        printc.note("This command will create all the configuration files needed by Odatix in the current directory.", script_name=script_name)
        printc.warning("This will overwrite any existing configuration files.", script_name=script_name)
        printc.say("Would you like to continue? ", end="", script_name=script_name)
        answer = ask_yes_no()
        if answer == False:
            return False
        success = OdatixSettings.init_path()
        if not success:
            return False
        if include_examples is None:
            printc.say("Would you like to add Odatix examples? ", end="", script_name=script_name)
            include_examples = ask_yes_no()
        if include_examples:
            success = OdatixSettings.init_examples()
            if not success:
                return False
        OdatixSettings.init_success(prog)
        return True

    @staticmethod
    def init_directory_nodialog(include_examples=None, prog="", silent=False):
        """
        Initialize the configuration directory without user interaction.

        Args:
            include_examples (bool | None): Whether to include example files.
            prog (str): Program name for success message.

        Returns:
            bool: True if initialization is successful, False otherwise.
        """
        success = OdatixSettings.init_path()
        if not success:
            return False
        if include_examples:
            success = OdatixSettings.init_examples()
            if not success:
                return False
        if not silent:
            OdatixSettings.init_success(prog)
        return True

    @staticmethod
    def init_success(prog=""):
        """
        Display a success message after directory initialization.

        Args:
            prog (str): Program name for the message.
        """
        printc.green("Your directory can now be used by Odatix!", script_name=script_name)
        printc.say("Run ", end="", script_name=script_name)
        printc.bold(prog + " -h", end="")
        print(" to get a list of useful commands")

    def to_dict(self):
        """
        Returns a dictionary representation of the current settings.
        """
        return {
            # Main paths
            "arch_path": self.arch_path,
            "sim_path": self.sim_path,
            "target_path": self.target_path,

            # Settings files
            "clean_settings_file": self.clean_settings_file,
            "simulation_settings_file": self.simulation_settings_file,
            "fmax_synthesis_settings_file": self.fmax_synthesis_settings_file,
            "custom_freq_synthesis_settings_file": self.custom_freq_synthesis_settings_file,

            # Work directories
            "work_path": self.work_path,
            "simulation_work_path": self.simulation_work_path,
            "fmax_synthesis_work_path": self.fmax_synthesis_work_path,
            "custom_freq_synthesis_work_path": self.custom_freq_synthesis_work_path,

            # Results   
            "result_path": self.result_path,
            "use_benchmark": self.use_benchmark,
            "benchmark_file": self.benchmark_file,
        }

    @staticmethod
    def from_dict(settings_dict):
        """
        Create an OdatixSettings object from a dictionary.

        Args:
            settings_dict (dict): Dictionary containing settings.
        Returns:
            OdatixSettings: The created settings object.
        """
        settings = OdatixSettings()
        for key, value in settings_dict.items():
            try:
                setattr(settings, key, value)
            except Exception as e:
                printc.warning(f"Key '{key}' is not a valid attribute of OdatixSettings: {str(e)}. Skipping key.", script_name=script_name)
        return settings

    def save_to_file(self, settings_filename=DEFAULT_SETTINGS_FILE):
        """
        Save the current settings to a YAML file.

        Args:
            settings_filename (str): Path to the settings file.
        """
        settings = self.to_dict()
        OdatixSettings.save_dict_to_file(settings, settings_filename)
    
    @staticmethod
    def save_dict_to_file(settings, settings_filename=DEFAULT_SETTINGS_FILE):
        try:
            from ruamel.yaml import YAML, CommentedMap, CommentedSeq
        except ImportError:
            printc.error("The 'ruamel.yaml' package is required to save settings files. Please install it via 'pip install ruamel.yaml'", script_name=script_name)
            return

        yaml_obj = YAML()

        data = CommentedMap()

        data.yaml_set_start_comment(
f"""##############################################
# Odatix Workspace Settings
##############################################

# Default workspace settings can be overridden here.

# This file was generated by Odatix GUI {motd.read_version()} on {datetime.now().strftime("%Y-%m-%d %H:%M:%S")}.
# You can still modify manually this file as needed.
"""
        )

        data['workspace'] = "odatix"
        data.yaml_set_comment_before_after_key('workspace', before="\n", after="\n")

        # Main paths
        arch_path = settings.get('arch_path')
        if arch_path is not None and arch_path != "":
            data['arch_path'] = arch_path
        sim_path = settings.get('sim_path')
        if sim_path is not None and sim_path != "":
            data['sim_path'] = sim_path
        target_path = settings.get('target_path')
        if target_path is not None and target_path != "":
            data['target_path'] = target_path

        if arch_path is not None and arch_path != "":
            data.yaml_set_comment_before_after_key('arch_path', before="\nMain paths")
        elif sim_path is not None and sim_path != "":
            data.yaml_set_comment_before_after_key('sim_path', before="\nMain paths")
        elif target_path is not None and target_path != "":
            data.yaml_set_comment_before_after_key('target_path', before="\nMain paths")

        # Settings files
        clean_settings_file = settings.get('clean_settings_file')
        if clean_settings_file is not None and clean_settings_file != "":
            data['clean_settings_file'] = clean_settings_file
        simulation_settings_file = settings.get('simulation_settings_file')
        if simulation_settings_file is not None and simulation_settings_file != "":
            data['simulation_settings_file'] = simulation_settings_file
        fmax_synthesis_settings_file = settings.get('fmax_synthesis_settings_file')
        if fmax_synthesis_settings_file is not None and fmax_synthesis_settings_file != "":
            data['fmax_synthesis_settings_file'] = fmax_synthesis_settings_file
        custom_freq_synthesis_settings_file = settings.get('custom_freq_synthesis_settings_file')
        if custom_freq_synthesis_settings_file is not None and custom_freq_synthesis_settings_file != "":
            data['custom_freq_synthesis_settings_file'] = custom_freq_synthesis_settings_file

        if clean_settings_file is not None and clean_settings_file != "":
            data.yaml_set_comment_before_after_key('clean_settings_file', before="\nSettings files")
        elif simulation_settings_file is not None and simulation_settings_file != "":
            data.yaml_set_comment_before_after_key('simulation_settings_file', before="\nSettings files")
        elif fmax_synthesis_settings_file is not None and fmax_synthesis_settings_file != "":
            data.yaml_set_comment_before_after_key('fmax_synthesis_settings_file', before="\nSettings files")
        elif custom_freq_synthesis_settings_file is not None and custom_freq_synthesis_settings_file != "":
            data.yaml_set_comment_before_after_key('custom_freq_synthesis_settings_file', before="\nSettings files")

        # Work directories
        work_path = settings.get('work_path')
        if work_path is not None and work_path != "":
            data['work_path'] = work_path
        simulation_work_path = settings.get('simulation_work_path')
        if simulation_work_path is not None and simulation_work_path != "":
            data['simulation_work_path'] = simulation_work_path
        fmax_synthesis_work_path = settings.get('fmax_synthesis_work_path')
        if fmax_synthesis_work_path is not None and fmax_synthesis_work_path != "":
            data['fmax_synthesis_work_path'] = fmax_synthesis_work_path
        custom_freq_synthesis_work_path = settings.get('custom_freq_synthesis_work_path')
        if custom_freq_synthesis_work_path is not None and custom_freq_synthesis_work_path != "":
            data['custom_freq_synthesis_work_path'] = custom_freq_synthesis_work_path

        if work_path is not None and work_path != "":
            data.yaml_set_comment_before_after_key('work_path', before="\nWork directories")
        elif simulation_work_path is not None and simulation_work_path != "":
            data.yaml_set_comment_before_after_key('simulation_work_path', before="\nWork directories")
        elif fmax_synthesis_work_path is not None and fmax_synthesis_work_path != "":
            data.yaml_set_comment_before_after_key('fmax_synthesis_work_path', before="\nWork directories")
        elif custom_freq_synthesis_work_path is not None and custom_freq_synthesis_work_path != "":
            data.yaml_set_comment_before_after_key('custom_freq_synthesis_work_path', before="\nWork directories")

        # Results
        result_path = settings.get('result_path')
        if result_path is not None and result_path != "":
            data['result_path'] = result_path
        use_benchmark = settings.get('use_benchmark')
        if use_benchmark is not None and use_benchmark != "":
            data['use_benchmark'] = use_benchmark
        benchmark_file = settings.get('benchmark_file')
        if benchmark_file is not None and benchmark_file != "":
            data['benchmark_file'] = benchmark_file

        if result_path is not None and result_path != "":
            data.yaml_set_comment_before_after_key('result_path', before="\nResults")
        elif use_benchmark is not None and use_benchmark != "":
            data.yaml_set_comment_before_after_key('use_benchmark', before="\nResults")
        elif benchmark_file is not None and benchmark_file != "":
            data.yaml_set_comment_before_after_key('benchmark_file', before="\nResults")

        with open(settings_filename, "w") as f:
            yaml_obj.dump(data, f)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import dash
from dash import html
import odatix.gui.navigation as navigation

dash.register_page(
    __name__,
    name='PageNotFound'
)

######################################
# Layout
######################################

layout = html.Div(
    children=[
        html.Div(
            children=[
                html.H1("404"),
                html.H2("Oops, page not found!")
            ],
            className="tile center"
        ),
    ],
    className="page-content",
    id="p404-container",
    style={
        "width": "100%", 
        "minHeight": f"calc(100vh - {navigation.top_bar_height})",
        "display": "flex",  
        "flexDirection": "column",
        "alignItems": "center",
    },
)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import re
import dash
from dash import dcc, html
from dash.dependencies import Input, Output
import plotly.graph_objs as go
import pandas as pd

import odatix.explorer.legend as legend

from odatix.lib.utils import safe_df_append

def add_trace_to_lines_fig(fig, x_values, y_values, mode, architecture, frequency, targets, target, selected_metric_display, unit, color_id, symbol_id, toggle_legendgroup, toggle_connect_gaps):
  fig.add_trace(
    go.Scatter(
      x=x_values,
      y=y_values,
      mode=mode,
      line=dict(dash="dot") if mode == "lines+markers" else None,
      marker=dict(
        size=10,
        color=legend.get_color(color_id),
        symbol=legend.get_marker_symbol(symbol_id),
      ),
      name=f"{architecture} @ {frequency}" if frequency else architecture,
      customdata=targets,
      legendgroup=target,
      legendgrouptitle_text=str(target) if toggle_legendgroup else None,
      connectgaps=True if toggle_connect_gaps else False,
      hovertemplate="<br>".join(
        [
          "Architecture: %{fullData.name}",
          "Configuration: %{x}",
          "Target: %{customdata}",
          selected_metric_display + ": %{y} " + unit,
          "<extra></extra>",
        ]
      ),
    )
  )

def add_trace_to_columns_fig(fig, x_values, y_values, mode, architecture, frequency, targets, target, selected_metric_display, unit, color_id, pattern_id, toggle_legendgroup):
  fig.add_trace(
    go.Bar(
      x=x_values,
      y=y_values,
      marker=dict(
        color=legend.get_color(color_id),
        pattern=dict(
          shape=legend.get_pattern(pattern_id)
        )
      ),
      name=f"{architecture} @ {frequency}" if frequency else architecture,
      customdata=targets,
      legendgroup=target,
      legendgrouptitle_text=str(target) if toggle_legendgroup else None,
      hovertemplate="<br>".join(
        [
          "Architecture: %{fullData.name}",
          "Configuration: %{x}",
          "Target: %{customdata}",
          selected_metric_display + ": %{y} " + unit,
          "<extra></extra>",
        ]
      ),
    )
  )

def add_trace_to_scatter_fig(fig, x_values, y_values, mode, architecture, frequencies, frequency, config_names, targets, target, selected_metric_x_display, selected_metric_y_display, unit_x, unit_y, color_id, symbol_id, toggle_lines, toggle_legendgroup):
  fig.add_trace(
    go.Scatter(
      x=x_values,
      y=y_values,
      mode=mode,
      line=dict(dash="dot") if toggle_lines else None,
      marker=dict(
        size=10,
        color=legend.get_color(color_id ),
        symbol=legend.get_marker_symbol(symbol_id),
      ),
      name=f"{architecture} @ {frequency}" if frequency else architecture,
      customdata=[list(a) for a in zip(targets, frequencies)],
      legendgroup=target,
      legendgrouptitle_text=str(target) if toggle_legendgroup else None,
      connectgaps=True,
      text=config_names,
      textposition="top center",
      hovertemplate="<br>".join(
        [
          "Architecture: %{fullData.name}",
          "Configuration: %{text}",
          "Target: %{customdata[0]}",
          "Frequency: %{customdata[1]}",
          selected_metric_x_display + ": %{x} " + unit_x,
          selected_metric_y_display + ": %{y} " + unit_y,
          "<extra></extra>",
        ]
      ),
    )
  )

def add_trace_to_scatter3d_fig(fig, x_values, y_values, z_values, mode, architecture, frequencies, frequency, config_names, targets, target, selected_metric_x_display, selected_metric_y_display, selected_metric_z_display, unit_x, unit_y, unit_z, color_id, symbol_id, toggle_lines, toggle_legendgroup):
  fig.add_trace(
    go.Scatter3d(
      x=x_values,
      y=y_values,
      z=z_values,
      mode=mode,
      line=dict(dash="dot") if toggle_lines else None,
      marker=dict(
        size=5,
        color=legend.get_color(color_id),
        symbol=legend.get_marker_symbol_3d(symbol_id),
      ),
      name=f"{architecture} @ {frequency}" if frequency else architecture,
      customdata=[list(a) for a in zip(targets, frequencies)],
      legendgroup=target,
      legendgrouptitle_text=str(target) if toggle_legendgroup else None,
      connectgaps=True,
      text=config_names,
      textposition="top center",
      hovertemplate="<br>".join(
        [
          "Architecture: %{fullData.name}",
          "Configuration: %{text}",
          "Target: %{customdata[0]}",
          "Frequency: %{customdata[1]}",
          selected_metric_x_display + ": %{x} " + unit_x,
          selected_metric_y_display + ": %{y} " + unit_y,
          selected_metric_z_display + ": %{z} " + unit_z,
          "<extra></extra>",
        ]
      ),
    )
  )

def add_trace_to_radar_fig(fig, theta_values, r_values, mode, architecture, frequency, targets, target, selected_metric_display, unit, color_id, symbol_id, toggle_legendgroup, toggle_connect_gaps):
  fig.add_trace(
  go.Scatterpolar(
    theta=theta_values,
    r=r_values,
    mode=mode,
    name=f"{architecture} @ {frequency}" if frequency else architecture,
    customdata=targets,
    legendgroup=target,
    legendgrouptitle_text=str(target) if toggle_legendgroup else None,
    connectgaps=True if toggle_connect_gaps else False,
    marker_size=10,
    marker_color=legend.get_color(color_id),
    marker_symbol=legend.get_marker_symbol(symbol_id),
    hovertemplate="<br>".join(
      [
        "Architecture: %{fullData.name}",
        "Configuration: %{theta}",
        "Target: %{customdata}",
        selected_metric_display + ": %{y} " + unit,
        "<extra></extra>",
      ]
    ),
  )
)

# Overview 

def make_figure_div(fig, filename, dl_format, page_wide=False, remove_zoom=False):
  page_wide_style = {"width": "100%", "height": "100%", "display": "inline-block", "vertical-align": "top"}
  normal_style = {"flex": "0 0 auto", "margin": "0px"}

  if fig is not None:
    to_remove = ["lasso", "select"]
    if remove_zoom:
      to_remove.append("zoom")
    return html.Div(
      [
        dcc.Graph(
          figure=fig,
          style={"width": "100%"},
          config={
            "displayModeBar": True,
            "displaylogo": False,
            "modeBarButtonsToRemove": to_remove,
            "toImageButtonOptions": {
              "format": dl_format,
              "scale": 3,
              "filename": filename
            },
          },
        )
      ],
      style=page_wide_style if page_wide else normal_style,
    )
  else:
    return html.Div([html.Div([], style={"width": "475px"})], style={"flex": "0 0 auto", "margin": "0px"})
